#!/usr/bin/env python3
import math
import os

# import matplotlib.ticker as ticker
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from collections import namedtuple
from scipy import linalg
from pathlib import Path
from math import atan
# import numba
# from numpy import arctan as atan

import artistools as at
import artistools.estimators
import artistools.nltepops
import artistools.nonthermal

# cgs units to match artis
EV = 1.6021772e-12  # in erg
H = 6.6260755e-27
ME = 9.1093897e-28
QE = 4.80325E-10
H_ionpot = 13.5979996 * EV
CLIGHT = 2.99792458e+10
PI = math.pi

use_collstrengths = False


def get_nntot(ions, ionpopdict):
    # total number density of all nuclei [cm^-3]
    nntot = 0.
    for Z, ionstage in ions:
        nntot += ionpopdict[(Z, ionstage)]
    return nntot


def get_nne(ions, ionpopdict):
    # number density of free electrons [cm-^3]
    nne = 0.
    for Z, ionstage in ions:
        charge = ionstage - 1
        assert(charge >= 0)
        nne += charge * ionpopdict[(Z, ionstage)]

    return nne


def get_Zbar(ions, ionpopdict):
    # number density-weighted average atomic number
    # i.e. protons per nucleus
    Zbar = 0.
    nntot = get_nntot(ions, ionpopdict)
    for Z, ionstage in ions:
        nnion = ionpopdict[(Z, ionstage)]
        Zbar += Z * nnion / nntot

    return Zbar


def get_Zboundbar(ions, ionpopdict):
    # number density-weighted average number of bound electrons per nucleus
    Zboundbar = 0.
    nntot = get_nntot(ions, ionpopdict)
    for Z, ionstage in ions:
        nnion = ionpopdict[(Z, ionstage)]
        Zboundbar += (Z - ionstage + 1) * nnion / nntot

    return Zboundbar


def get_nnetot(ions, ionpopdict):
    # total number density of electrons (free + bound) [cm-^3]
    # return get_Zbar(ions, ionpopdict) * get_nntot(ions, ionpopdict)
    nnetot = 0.
    for Z, ionstage in ions:
        nnetot += Z * ionpopdict[(Z, ionstage)]

    return nnetot


def read_binding_energies(modelpath='.'):
    collionfilename = at.firstexisting([os.path.join(modelpath, 'binding_energies.txt'),
                                        os.path.join(at.PYDIR, 'data', 'binding_energies.txt')])

    with open(collionfilename, "r") as f:
        nt_shells, n_z_binding = [int(x) for x in f.readline().split()]
        electron_binding = np.zeros((n_z_binding, nt_shells))

        for i in range(n_z_binding):
            electron_binding[i] = np.array([float(x) for x in f.readline().split()]) * EV

    return electron_binding


def get_electronoccupancy(atomic_number, ion_stage, nt_shells):
    # adapted from ARTIS code
    q = np.zeros(nt_shells)

    ioncharge = ion_stage - 1
    nbound = atomic_number - ioncharge  # number of bound electrons

    for electron_loop in range(nbound):
        if q[0] < 2:  # K 1s
            q[0] += 1
        elif(q[1] < 2):  # L1 2s
            q[1] += 1
        elif(q[2] < 2):  # L2 2p[1/2]
            q[2] += 1
        elif(q[3] < 4):  # L3 2p[3/2]
            q[3] += 1
        elif(q[4] < 2):  # M1 3s
            q[4] += 1
        elif(q[5] < 2):  # M2 3p[1/2]
            q[5] += 1
        elif(q[6] < 4):  # M3 3p[3/2]
            q[6] += 1
        elif ioncharge == 0:
            if q[9] < 2:  # N1 4s
                q[9] += 1
            elif q[7] < 4:  # M4 3d[3/2]
                q[7] += 1
            elif q[8] < 6:  # M5 3d[5/2]
                q[8] += 1
            else:
                print("Going beyond the 4s shell in NT calculation. Abort!\n")
        elif ioncharge == 1:
            if q[9] < 1:  # N1 4s
                q[9] += 1
            elif q[7] < 4:  # M4 3d[3/2]
                q[7] += 1
            elif q[8] < 6:  # M5 3d[5/2]
                q[8] += 1
            else:
                print("Going beyond the 4s shell in NT calculation. Abort!\n")
        elif(ioncharge > 1):
            if q[7] < 4:  # M4 3d[3/2]
                q[7] += 1
            elif q[8] < 6:  # M5 3d[5/2]
                q[8] += 1
            else:
                print("Going beyond the 4s shell in NT calculation. Abort!\n")
    return q


def get_mean_binding_energy(atomic_number, ion_stage, electron_binding, ionpot_ev):
    # LJS: this came from ARTIS and I'm not sure what this actually is - inverse binding energy? electrons per erg?
    n_z_binding, nt_shells = electron_binding.shape
    q = get_electronoccupancy(atomic_number, ion_stage, nt_shells)

    total = 0.0
    for electron_loop in range(nt_shells):
        electronsinshell = q[electron_loop]
        if ((electronsinshell) > 0):
            use2 = electron_binding[atomic_number - 1][electron_loop]
            use3 = ionpot_ev * EV
        if (use2 <= 0):
            use2 = electron_binding[atomic_number - 1][electron_loop - 1]
            # to get total += electronsinshell/electron_binding[get_element(element)-1][electron_loop-1];
            # set use3 = 0.
            if (electron_loop != 8):
                # For some reason in the Lotz data, this is no energy for the M5 shell before Ni. So if the complaint
                # is for 8 (corresponding to that shell) then just use the M4 value
                print("WARNING: I'm trying to use a binding energy when I have no data. "
                      f"element {atomic_number} ionstage {ion_stage}\n")
                assert(electron_loop == 8)
                # print("Z = %d, ion_stage = %d\n", get_element(element), get_ionstage(element, ion));
        if (use2 < use3):
            total += electronsinshell / use3
        else:
            total += electronsinshell / use2
        # print("total total)

    return total


def get_mean_binding_energy_alt(atomic_number, ion_stage, electron_binding, ionpot_ev):
    # LJS: this should be mean binding energy [erg] per electron
    n_z_binding, nt_shells = electron_binding.shape
    q = get_electronoccupancy(atomic_number, ion_stage, nt_shells)

    total = 0.0
    ecount = 0
    for electron_loop in range(nt_shells):
        electronsinshell = q[electron_loop]
        ecount += electronsinshell
        if ((electronsinshell) > 0):
            use2 = electron_binding[atomic_number - 1][electron_loop]
            use3 = ionpot_ev * EV
        if (use2 <= 0):
            use2 = electron_binding[atomic_number - 1][electron_loop - 1]
            # to get total += electronsinshell/electron_binding[get_element(element)-1][electron_loop-1];
            # set use3 = 0.
            if (electron_loop != 8):
                # For some reason in the Lotz data, this is no energy for the M5 shell before Ni. So if the complaint
                # is for 8 (corresponding to that shell) then just use the M4 value
                print("WARNING: I'm trying to use a binding energy when I have no data. "
                      f"element {atomic_number} ionstage {ion_stage}\n")
                assert(electron_loop == 8)
                # print("Z = %d, ion_stage = %d\n", get_element(element), get_ionstage(element, ion));
        if (use2 < use3):
            total += electronsinshell * use3
        else:
            total += electronsinshell * use2

    assert ecount == (atomic_number - ion_stage + 1)

    return total / ecount


def get_lotz_xs_ionisation(atomic_number, ion_stage, electron_binding, ionpot_ev, en_ev):
    # Axelrod 1980 Eq 3.38

    en_erg = en_ev * EV
    gamma = en_erg / (ME * CLIGHT ** 2) + 1
    beta = math.sqrt(1. - 1. / (gamma ** 2))
    # beta = 0.99
    # print(f'{gamma=} {beta=}')

    n_z_binding, nt_shells = electron_binding.shape
    q = get_electronoccupancy(atomic_number, ion_stage, nt_shells)

    part_sigma = 0.0
    for electron_loop in range(nt_shells):
        electronsinshell = q[electron_loop]
        if ((electronsinshell) > 0):
            use2 = electron_binding[atomic_number - 1][electron_loop]
            use3 = ionpot_ev * EV
        if (use2 <= 0):
            use2 = electron_binding[atomic_number - 1][electron_loop - 1]
            # to get total += electronsinshell/electron_binding[get_element(element)-1][electron_loop-1];
            # set use3 = 0.
            if (electron_loop != 8):
                # For some reason in the Lotz data, this is no energy for the M5 shell before Ni. So if the complaint
                # is for 8 (corresponding to that shell) then just use the M4 value
                print("WARNING: I'm trying to use a binding energy when I have no data. "
                      f"element {atomic_number} ionstage {ion_stage}\n")
                assert(electron_loop == 8)
                # print("Z = %d, ion_stage = %d\n", get_element(element), get_ionstage(element, ion));

        if (use2 < use3):
            p = use3
        else:
            p = use2

        if 0.5 * beta ** 2 * ME * CLIGHT ** 2 > p:
            part_sigma += electronsinshell / p * (
                (math.log(beta ** 2 * ME * CLIGHT ** 2 / 2. / p) - math.log10(1 - beta ** 2) - beta ** 2))

    Aconst = 1.33e-14 * EV * EV
    # me is electron mass
    sigma = 2 * Aconst / (beta ** 2) / ME / (CLIGHT ** 2) * part_sigma
    assert(sigma >= 0)
    return sigma


def lossfunction(energy_ev, nne_cgs):
    # free-electron plasma loss rate (as in Kozma & Fransson 1992)
    # - dE / dX [eV / cm]
    # returns a positive number

    # return math.log(energy_ev) / energy_ev
    if nne_cgs == 0:
        return 0.

    nne = nne_cgs
    energy = energy_ev * EV  # convert eV to erg

    # omegap = math.sqrt(4 * math.pi * nne_selected_cgs * pow(QE, 2) / ME)
    omegap = 5.64e4 * math.sqrt(nne_cgs)
    zetae = H * omegap / 2 / math.pi

    if energy_ev > 14:
        assert 2 * energy > zetae
        lossfunc = nne * 2 * math.pi * QE ** 4 / energy * math.log(2 * energy / zetae)
    else:
        v = math.sqrt(2 * energy / ME)  # velocity in m/s
        eulergamma = 0.577215664901532
        lossfunc = nne * 2 * math.pi * QE ** 4 / energy * math.log(ME * pow(v, 3) / (eulergamma * pow(QE, 2) * omegap))

    # lossfunc is now [erg / cm]
    return lossfunc / EV  # return as [eV / cm]


def Psecondary(e_p, ionpot_ev, J, e_s=-1, epsilon=-1):
    # probability distribution function for secondaries energy e_s [eV] (or equivalently the energy loss of
    # the primary electron epsilon [eV]) given a primary energy e_p [eV] for an impact ionisation event

    assert e_s >= 0 or epsilon >= 0
    # if e_p < I:
    #     return 0.
    #
    if e_s < 0:
        e_s = epsilon - ionpot_ev
    if epsilon < 0:
        epsilon = e_s + ionpot_ev

    #
    # if epsilon < I:
    #     return 0.
    # if e_s < 0:
    #     return 0.
    # if e_s > e_p - I:
    #     return 0.
    # if e_s > e_p:
    #     return 0.

    # test case: constant, always below ionisation
    # Psecondary_e_s_max = 1. / J / 2.
    # return 1. / Psecondary_e_s_max if (e_s < Psecondary_e_s_max) else 0.

    return 1. / J / atan((e_p - ionpot_ev) / 2. / J) / (1 + ((e_s / J) ** 2))


def get_J(Z, ionstage, ionpot_ev):
    # returns an energy in eV
    # values from Opal et al. 1971 as applied by Kozma & Fransson 1992
    if (ionstage == 1):
        if (Z == 2):  # He I
            return 15.8
        elif (Z == 10):  # Ne I
            return 24.2
        elif (Z == 18):  # Ar I
            return 10.0

    return 0.6 * ionpot_ev
    # return 2.0 * ionpot_ev
    # return 16.0 * ionpot_ev
    # return 500


def ar_xs(energy_ev, ionpot_ev, A, B, C, D):
    u = energy_ev / ionpot_ev
    if u <= 1:
        return 0

    return 1e-14 * (
        A * (1 - 1 / u) + B * pow((1 - 1 / u), 2) + C * math.log(u) + D * math.log(u) / u) / (u * pow(ionpot_ev, 2))


def get_arxs_array_shell(arr_enev, shell):
    ar_xs_array = np.array(
        [ar_xs(energy_ev, shell.ionpot_ev, shell.A, shell.B, shell.C, shell.D) for energy_ev in arr_enev])

    return ar_xs_array


def get_arxs_array_ion(arr_enev, dfcollion, Z, ionstage):
    ar_xs_array = np.zeros(len(arr_enev))
    dfcollion_thision = dfcollion.query('Z == @Z and ionstage == @ionstage')
    for index, shell in dfcollion_thision.iterrows():
        ar_xs_array += get_arxs_array_shell(arr_enev, shell)

    return ar_xs_array


def get_xs_excitation(en_ev, row):
    A_naught_squared = 2.800285203e-17  # Bohr radius squared in cm^2

    coll_str = row.collstr
    epsilon_trans = row.epsilon_trans_ev * EV
    epsilon_trans_ev = row.epsilon_trans_ev

    if en_ev < epsilon_trans_ev:
        return 0.

    if coll_str >= 0 and use_collstrengths:
        # collision strength is available, so use it
        # Li et al. 2012 equation 11
        constantfactor = pow(H_ionpot, 2) / row.lower_g * coll_str * math.pi * A_naught_squared

        return constantfactor * (en_ev * EV) ** -2

    elif not row.forbidden:

        nu_trans = epsilon_trans / H
        g = row.upper_g / row.lower_g
        fij = g * ME * pow(CLIGHT, 3) / (8 * pow(QE * nu_trans * math.pi, 2)) * row.A
        # permitted E1 electric dipole transitions

        g_bar = 0.2

        A = 0.28
        B = 0.15

        prefactor = 45.585750051
        # Eq 4 of Mewe 1972, possibly from Seaton 1962?
        constantfactor = prefactor * A_naught_squared * pow(H_ionpot / epsilon_trans, 2) * fij

        U = en_ev / epsilon_trans_ev
        g_bar = A * np.log(U) + B

        return constantfactor * g_bar / U

    return 0.


def get_xs_excitation_vector(engrid, row):
    A_naught_squared = 2.800285203e-17  # Bohr radius squared in cm^2
    npts = len(engrid)
    xs_excitation_vec = np.empty(npts)

    coll_str = row.collstr
    epsilon_trans = row.epsilon_trans_ev * EV
    epsilon_trans_ev = row.epsilon_trans_ev

    startindex = get_energyindex_gteq(en_ev=epsilon_trans_ev, engrid=engrid)
    xs_excitation_vec[:startindex] = 0.

    if coll_str >= 0 and use_collstrengths:
        # collision strength is available, so use it
        # Li et al. 2012 equation 11
        constantfactor = pow(H_ionpot, 2) / row.lower_g * coll_str * math.pi * A_naught_squared

        xs_excitation_vec[startindex:] = constantfactor * (engrid[startindex:] * EV) ** -2

    elif not row.forbidden:

        nu_trans = epsilon_trans / H
        g = row.upper_g / row.lower_g
        fij = g * ME * pow(CLIGHT, 3) / (8 * pow(QE * nu_trans * math.pi, 2)) * row.A
        # permitted E1 electric dipole transitions

        g_bar = 0.2

        A = 0.28
        B = 0.15

        prefactor = 45.585750051
        # Eq 4 of Mewe 1972, possibly from Seaton 1962?
        constantfactor = prefactor * A_naught_squared * pow(H_ionpot / epsilon_trans, 2) * fij

        U = engrid[startindex:] / epsilon_trans_ev
        g_bar = A * np.log(U) + B

        xs_excitation_vec[startindex:] = constantfactor * g_bar / U
        for j, energy_ev in enumerate(engrid):
            energy = energy_ev * EV
            if (energy >= epsilon_trans):
                U = energy / epsilon_trans
                g_bar = A * math.log(U) + B
                xs_excitation_vec[j] = constantfactor * g_bar / U
    else:
        xs_excitation_vec[startindex:] = 0.

    return xs_excitation_vec


def read_colliondata(collionfilename='collion.txt', modelpath=Path(at.PYDIR, 'data')):

    collionrow = namedtuple('collionrow', ['Z', 'nelec', 'n', 'l', 'ionpot_ev', 'A', 'B', 'C', 'D'])

    nrows = -1
    with open(Path(modelpath, collionfilename), 'r') as collionfile:
        nrows = int(collionfile.readline().strip())
        # print(f'Collionfile: expecting {nrows} rows')
        dfcollion = pd.read_csv(
            collionfile, delim_whitespace=True, header=None, names=collionrow._fields)

    # assert len(dfcollion) == nrows  # artis enforces this, but last 10 rows were not inportant anyway (high ionized Ni)
    dfcollion.eval('ionstage = Z - nelec + 1', inplace=True)

    return dfcollion


def calculate_nt_frac_excitation(engrid, dftransitions, yvec, deposition_density_ev):
    # Kozma & Fransson equation 4, but summed over all transitions for given ion
    # integral in Kozma & Fransson equation 9
    deltaen = engrid[1] - engrid[0]
    npts = len(engrid)

    xs_excitation_vec_sum_alltrans = np.zeros(npts)

    for _, row in dftransitions.iterrows():
        nnlevel = row.lower_pop
        xs_excitation_vec_sum_alltrans += nnlevel * row.epsilon_trans_ev * get_xs_excitation_vector(engrid, row)

    return np.dot(xs_excitation_vec_sum_alltrans, yvec) * deltaen / deposition_density_ev


def get_energyindex_lteq(en_ev, engrid):
    # find energy bin lower boundary is less than or equal to search value
    # assert en_ev >= engrid[0]
    deltaen = engrid[1] - engrid[0]
    # assert en_ev < (engrid[-1] + deltaen)

    index = math.floor((en_ev - engrid[0]) / deltaen)

    if index < 0:
        return 0
    elif (index > len(engrid) - 1):
        return len(engrid) - 1
    else:
        return index


def get_energyindex_gteq(en_ev, engrid):
    # find energy bin lower boundary is greater than or equal to search value
    deltaen = engrid[1] - engrid[0]

    index = math.ceil((en_ev - engrid[0]) / deltaen)

    if index < 0:
        return 0
    elif (index > len(engrid) - 1):
        return len(engrid) - 1
    else:
        return index


def calculate_N_e(energy_ev, engrid, ions, ionpopdict, dfcollion, yvec, dftransitions, noexcitation):
    # Kozma & Fransson equation 6.
    # Something related to a number of electrons, needed to calculate the heating fraction in equation 3
    # not valid for energy > E_0
    if energy_ev == 0.:
        return 0.

    N_e = 0.

    deltaen = engrid[1] - engrid[0]

    for Z, ionstage in ions:
        N_e_ion = 0.
        nnion = ionpopdict[(Z, ionstage)]

        if not noexcitation:
            if (Z, ionstage) in dftransitions:
                for _, row in dftransitions[(Z, ionstage)].iterrows():
                    nnlevel = row.lower_pop
                    epsilon_trans_ev = row.epsilon_trans_ev
                    if energy_ev + epsilon_trans_ev >= engrid[0]:
                        i = get_energyindex_lteq(en_ev=energy_ev + epsilon_trans_ev, engrid=engrid)
                        N_e_ion += (nnlevel / nnion) * yvec[i] * get_xs_excitation(engrid[i], row)
                        # enbelow = engrid[i]
                        # enabove = engrid[i + 1]
                        # x = (energy_ev - enbelow) / (enabove - enbelow)
                        # yvecinterp = (1 - x) * yvec[i] + x * yvec[i + 1]
                        # N_e_ion += (nnlevel / nnion) * yvecinterp * get_xs_excitation(energy_ev + epsilon_trans_ev, row)

        dfcollion_thision = dfcollion.query('Z == @Z and ionstage == @ionstage', inplace=False)

        for index, shell in dfcollion_thision.iterrows():
            ionpot_ev = shell.ionpot_ev

            enlambda = min(engrid[-1] - energy_ev, energy_ev + ionpot_ev)
            J = get_J(shell.Z, shell.ionstage, ionpot_ev)

            ar_xs_array = at.nonthermal.get_arxs_array_shell(engrid, shell)

            # integral from ionpot to enlambda
            integral1startindex = get_energyindex_lteq(en_ev=ionpot_ev, engrid=engrid)
            integral2stopindex = get_energyindex_lteq(en_ev=enlambda, engrid=engrid)

            for j in range(integral1startindex, integral2stopindex + 1):
                endash = engrid[j]
                k = get_energyindex_lteq(en_ev=energy_ev + endash, engrid=engrid)
                N_e_ion += deltaen * yvec[k] * ar_xs_array[k] * Psecondary(e_p=engrid[k], epsilon=endash, ionpot_ev=ionpot_ev, J=J)

                # interpolate the y value
                # xs = ar_xs(energy_ev + endash, shell.ionpot_ev, shell.A, shell.B, shell.C, shell.D)
                # enbelow = engrid[k]
                # enabove = engrid[k + 1]
                # x = (energy_ev - enbelow) / (enabove - enbelow)
                # yvecinterp = (1 - x) * yvec[k] + x * yvec[k + 1]
                # N_e_ion += deltaen * yvecinterp * xs * Psecondary(e_p=energy_ev + endash, epsilon=endash, ionpot_ev=ionpot_ev, J=J)

            # integral from 2E + I up to E_max
            integral2startindex = get_energyindex_lteq(en_ev=2 * energy_ev + ionpot_ev, engrid=engrid)
            N_e_ion += deltaen * sum([
                yvec[j] * ar_xs_array[j] *
                Psecondary(e_p=engrid[j], epsilon=energy_ev + ionpot_ev, ionpot_ev=ionpot_ev, J=J)
                for j in range(integral2startindex, len(engrid))])

        N_e += nnion * N_e_ion

    # source term not here because it should be zero at the low end anyway

    return N_e


def calculate_frac_heating(
        engrid, ions, ionpopdict, dfcollion, dftransitions, yvec, nne, nnetot, deposition_density_ev, noexcitation):
    # Kozma & Fransson equation 8

    frac_heating = 0.
    E_0 = engrid[0]

    deltaen = engrid[1] - engrid[0]
    frac_heating += deltaen / deposition_density_ev * sum([
        lossfunction(en_ev, nne, nnetot, ions=ions, ionpopdict=ionpopdict) * yvec[i]
        for i, en_ev in enumerate(engrid)])

    frac_heating_E_0_part = (
        E_0 * yvec[0] * lossfunction(E_0, nne, nnetot, ions=ions, ionpopdict=ionpopdict) / deposition_density_ev)

    frac_heating += frac_heating_E_0_part

    print(f"            frac_heating E_0 * y * l(E_0) part: {frac_heating_E_0_part:.5f}")

    frac_heating_N_e = 0.
    npts_integral = math.ceil(E_0 / deltaen) * 10
    print(f'N_e npts_integral: {npts_integral}')
    arr_en, deltaen2 = np.linspace(0., E_0, num=npts_integral, retstep=True, endpoint=True)
    arr_en_N_e = [en_ev * calculate_N_e(en_ev, engrid, ions, ionpopdict, dfcollion,
                                        yvec, dftransitions, noexcitation=noexcitation) for en_ev in arr_en]
    frac_heating_N_e += 1. / deposition_density_ev * sum(arr_en_N_e) * deltaen2

    print(f"            frac_heating N_e part: {frac_heating_N_e:.5f}")

    frac_heating += frac_heating_N_e

    return frac_heating


def sfmatrix_add_excitation(engrid, dftransitions_ion, nnion, sfmatrix):
    deltaen = engrid[1] - engrid[0]
    for _, row in dftransitions_ion.iterrows():
        nnlevel = row.lower_pop
        epsilon_trans_ev = row.epsilon_trans_ev
        if epsilon_trans_ev >= engrid[0]:
            vec_xs_excitation_nnlevel_deltae = nnlevel * deltaen * get_xs_excitation_vector(engrid, row)
            xsstartindex = get_energyindex_lteq(en_ev=epsilon_trans_ev, engrid=engrid)

            for i, en in enumerate(engrid):
                stopindex = get_energyindex_lteq(en_ev=en + epsilon_trans_ev, engrid=engrid)

                startindex = i if i > xsstartindex else xsstartindex
                # for j in range(startindex, stopindex):
                sfmatrix[i, startindex:stopindex] += vec_xs_excitation_nnlevel_deltae[startindex:stopindex]

                # do the last bit separately because we're not using the full deltaen interval

                delta_en_actual = (en + epsilon_trans_ev - engrid[stopindex])
                sfmatrix[i, stopindex] += vec_xs_excitation_nnlevel_deltae[stopindex] * delta_en_actual / deltaen


def sfmatrix_add_ionization_shell(engrid, nnion, shell, sfmatrix):
    # this code has been optimised and is now an almost unreadable form, but it contains the terms
    # related to ionisation cross sections
    deltaen = engrid[1] - engrid[0]
    ionpot_ev = shell.ionpot_ev
    J = get_J(shell.Z, shell.ionstage, ionpot_ev)
    npts = len(engrid)

    ar_xs_array = at.nonthermal.get_arxs_array_shell(engrid, shell)

    if ionpot_ev <= engrid[0]:
        xsstartindex = 0
    else:
        xsstartindex = npts + 1
        for i in range(npts):
            if ar_xs_array[i] > 0.:
                xsstartindex = i
                break
        xsstartindex = get_energyindex_gteq(en_ev=ionpot_ev, engrid=engrid)

    # J * atan[(epsilon - ionpot_ev) / J] is the indefinite integral of
    # 1/(1 + (epsilon - ionpot_ev)^2/ J^2) d_epsilon
    # in Kozma & Fransson 1992 equation 4

    prefactors = [nnion * ar_xs_array[j] / atan((engrid[j] - ionpot_ev) / 2. / J) * deltaen for j in range(npts)]
    epsilon_uppers = [min((engrid[j] + ionpot_ev) / 2, engrid[j]) for j in range(npts)]
    int_eps_uppers = [atan((epsilon_upper - ionpot_ev) / J) for epsilon_upper in epsilon_uppers]

    # for the resulting arrays, use index j - i corresponding to energy endash - en
    epsilon_lowers1 = [max(engrid[j] - engrid[0], ionpot_ev) for j in range(npts)]
    int_eps_lowers1 = [atan((epsilon_lower - ionpot_ev) / J) for epsilon_lower in epsilon_lowers1]

    for i, en in enumerate(engrid):
        # endash ranges from en to SF_EMAX, but skip over the zero-cross section points
        jstart = max(i, xsstartindex)

        # KF 92 limit
        # at each endash, the integral in epsilon ranges from
        # epsilon_lower = max(endash - en, ionpot_ev)
        # epsilon_upper = min((endash + ionpot_ev) / 2, endash)]

        # integral/J of 1/[1 + (epsilon - ionpot_ev) / J] for epsilon = en + ionpot_ev
        for j in range(jstart, npts):
            # j is the matrix column index which corresponds to the piece of the
            # integral at y(E') where E' >= E and E' = engrid[j]

            if epsilon_lowers1[j - i] <= epsilon_uppers[j]:
                sfmatrix[i, j] += prefactors[j] * (int_eps_uppers[j] - int_eps_lowers1[j - i])

        if 2 * en + ionpot_ev < engrid[-1] + (engrid[1] - engrid[0]):
            secondintegralstartindex = get_energyindex_lteq(2 * en + ionpot_ev, engrid)
        else:
            secondintegralstartindex = npts + 1

        # endash ranges from 2 * en + ionpot_ev to SF_EMAX
        # at each endash, the integral in epsilon ranges from
        # epsilon_lower = en + ionpot_ev
        # epsilon_upper = min((endash + ionpot_ev) / 2, endash)]
        epsilon_lower2 = en + ionpot_ev
        for j in range(secondintegralstartindex, npts):
            if epsilon_lower2 <= epsilon_uppers[j]:
                int_eps_lower2 = atan((epsilon_lower2 - ionpot_ev) / J)
                sfmatrix[i, j] -= prefactors[j] * (int_eps_uppers[j] - int_eps_lower2)


def differentialsfmatrix_add_ionization_shell(engrid, nnion, shell, sfmatrix):
    # this code has been optimised and is now an almost unreadable form, but it is the contains the terms
    # related to ionisation cross sections
    delta_en = engrid[1] - engrid[0]
    ionpot_ev = shell.ionpot_ev
    J = get_J(shell.Z, shell.ionstage, ionpot_ev)
    npts = len(engrid)

    ar_xs_array = at.nonthermal.get_arxs_array_shell(engrid, shell)

    if ionpot_ev <= engrid[0]:
        xsstartindex = 0
    else:
        xsstartindex = get_energyindex_lteq(en_ev=ionpot_ev, engrid=engrid)

    oneoveratangrid = 1. / np.arctan((engrid - ionpot_ev) / 2. / J)

    epsilon_lower_a = ionpot_ev
    int_eps_lower_a = atan((epsilon_lower_a - ionpot_ev) / J)
    for i in range(xsstartindex, npts):
        en = engrid[i]

        # integral of xs_ion(e_p=en, epsilon) with epsilon from I to (I + E) / 2

        epsilon_upper = (ionpot_ev + en) / 2.
        if (epsilon_lower_a < epsilon_upper):
            # P_int = 0.
            # eps_npts = 1000
            # delta_eps = (epsilon_upper - epsilon_lower) / eps_npts
            # for j in range(eps_npts):
            #     epsilon = epsilon_lower + j * delta_eps
            #     P_int += Psecondary(e_p=en, epsilon=epsilon, ionpot_ev=ionpot_ev, J=J) * delta_eps
            #
            # J * atan[(epsilon - ionpot_ev) / J] is the indefinite integral of
            # 1/(1 + (epsilon - ionpot_ev)^2/ J^2) d_epsilon

            int_eps_upper = atan((epsilon_upper - ionpot_ev) / J)
            P_int = 1. / atan((en - ionpot_ev) / 2. / J) * (int_eps_upper - int_eps_lower_a)
            # if int_eps_lower == int_eps_upper and epsilon_upper != epsilon_lower:
            # if (abs(P_int2 / P_int - 1.) > 0.2):
            #     print("warning eps low high int low high", epsilon_lower, epsilon_upper, int_eps_lower, int_eps_upper)
            #     print(f'{P_int=:.2e} {P_int2=:.2e} Ratio: {P_int2 / P_int:.2f}')

            sfmatrix[i, i] += nnion * ar_xs_array[i] * P_int

        enlambda = min(engrid[-1] - en, en + ionpot_ev)
        epsilon_lower = ionpot_ev
        epsilon_upper = enlambda
        if (epsilon_lower < epsilon_upper):
            eps_npts = 100
            delta_eps = (epsilon_upper - epsilon_lower) / eps_npts
            prefactor = nnion / J / atan((en - ionpot_ev) / 2. / J) * delta_eps
            for j in range(eps_npts):
                epsilon = epsilon_lower + j * delta_eps
                i_enpluseps = get_energyindex_lteq(en + epsilon, engrid=engrid)
                sfmatrix[i, i_enpluseps] -= prefactor * ar_xs_array[i_enpluseps] / (
                    1 + (((epsilon - ionpot_ev) / J) ** 2))

            # j_lower = get_energyindex_lteq(en + epsilon_lower, engrid=engrid)
            # j_upper = get_energyindex_lteq(en + epsilon_upper, engrid=engrid)
            # if (j_lower < j_upper):
            #     delta_eps = (epsilon_upper - epsilon_lower) / (j_upper - j_lower)
            #     prefactor = nnion / J / atan((en - ionpot_ev) / 2. / J) * delta_eps
            #     print(J, atan((en - ionpot_ev) / 2. / J))
            #     assert not math.isnan(prefactor)
            #     assert not math.isinf(prefactor)
            #     # for j in range(j_lower, j_upper):
            #     #     en_plus_epsilon = engrid[j]
            #     #     epsilon = en_plus_epsilon - en
            #     #     sfmatrix[i, j] -= prefactor * ar_xs_array[j] / (1 + (((epsilon - ionpot_ev) / J) ** 2))
            #     sfmatrix[i, j_lower:j_upper] -= prefactor * ar_xs_array[j_lower:j_upper] / (1 + ((((engrid[j_lower:j_upper] - en) - ionpot_ev) / J) ** 2))

        if (2 * en + ionpot_ev) < engrid[-1]:
            epsilon = en + ionpot_ev
            prefactor = nnion / J / (1 + (((epsilon - ionpot_ev) / J) ** 2)) * delta_en
            i_endash_lower = get_energyindex_lteq(2 * en + ionpot_ev, engrid)
            # for j in range(i_endash_lower, npts):
            #     sfmatrix[i, j] -= prefactor * ar_xs_array[j] * oneoveratangrid[j]
            sfmatrix[i, i_endash_lower:] -= prefactor * ar_xs_array[i_endash_lower:] * oneoveratangrid[i_endash_lower:]


def solve_spencerfano_differentialform(
        ions, ionpopdict, dfpops, nne, deposition_density_ev, engrid, sourcevec, dfcollion, args,
        adata=None, noexcitation=False):

    deltaen = engrid[1] - engrid[0]
    npts = len(engrid)
    nnetot = get_nnetot(ions, ionpopdict)

    print(f'\nSetting up differential-form Spencer-Fano equation with {npts} energy points'
          f' from {engrid[0]} to {engrid[-1]} eV...')

    E_init_ev = np.dot(engrid, sourcevec) * deltaen
    print(f'    E_init: {E_init_ev:7.2f} eV/s/cm3')

    constvec = np.zeros(npts)
    for i in range(npts):
        constvec[i] += sourcevec[i]

    lossfngrid = np.array([lossfunction(en, nne, nnetot, ions=ions, ionpopdict=ionpopdict) for en in engrid])

    sfmatrix = np.zeros((npts, npts))
    for i in range(npts):
        en = engrid[i]

        # - d/dE(y[E] * lossfn[E]) = - dy/dE(E) * lossfn(E) - y(E) * dlossfn/dE

        # - dy/dE(E) * lossfn(E) = - (y(E + deltaE) - y(E) ) / deltaen * lossfunction(E)

        # - ( - y(E) * lossfunction)
        sfmatrix[i, i] += lossfngrid[i] / deltaen

        if i + 1 < npts:
            # - y(E + deltaE) * lossfunction
            sfmatrix[i, i + 1] -= lossfngrid[i] / deltaen

        # - y(E) * dlossfn/dE
        sfmatrix[i, i] -= (lossfunction(en + deltaen, nne, nnetot, ions=ions, ionpopdict=ionpopdict) - lossfngrid[i]) / deltaen

    dftransitions = {}
    for Z, ionstage in ions:
        nnion = ionpopdict[(Z, ionstage)]
        print(f'  including Z={Z} ion_stage {ionstage} ({at.get_ionstring(Z, ionstage)}) nnion={nnion:.2e} ionisation')
        dfcollion_thision = dfcollion.query('Z == @Z and ionstage == @ionstage', inplace=False)
        # print(dfcollion_thision)

        for index, shell in dfcollion_thision.iterrows():
            # assert shell.ionpot_ev >= engrid[0]
            if shell.ionpot_ev < engrid[0]:
                print(f'  WARNING: first energy point at {engrid[0]} eV is above shell ionpot {shell.ionpot_ev} eV')
            differentialsfmatrix_add_ionization_shell(engrid, nnion, shell, sfmatrix)

        assert noexcitation

    print()
    lu_and_piv = linalg.lu_factor(sfmatrix, overwrite_a=False)
    yvec_reference = linalg.lu_solve(lu_and_piv, constvec, trans=0)
    yvec = yvec_reference * deposition_density_ev / E_init_ev

    return yvec, dftransitions


def get_nne_nt(engrid, yvec):
    # oneovervelocity = np.sqrt(9.10938e-31 / 2 / engrid / 1.60218e-19) / 100  # in s/cm
    # enovervelocity = engrid * oneovervelocity
    # en_tot = np.dot(yvec, enovervelocity) * (engrid[1] - engrid[0])
    nne_nt = 0.
    deltaen = (engrid[1] - engrid[0])
    for i, en in enumerate(engrid):
        # oneovervelocity = np.sqrt(9.10938e-31 / 2 / en / 1.60218e-19) / 100.
        velocity = np.sqrt(2 * en * 1.60218e-19 / 9.10938e-31) * 100.  # cm/s
        nne_nt += yvec[i] / velocity * deltaen

    return nne_nt


def analyse_ntspectrum(
        engrid, yvec, ions, ionpopdict, nntot, nne, deposition_density_ev, dfcollion, dftransitions,
        noexcitation, modelpath):

    deltaen = engrid[1] - engrid[0]

    frac_ionization = 0.
    frac_excitation = 0.
    frac_ionization_ion = {}
    frac_excitation_ion = {}
    gamma_nt = {}

    electron_binding = read_binding_energies(modelpath)

    Zbar = get_Zbar(ions, ionpopdict)
    nnetot = get_nnetot(ions, ionpopdict)

    Aconst = 1.33e-14 * EV * EV

    for Z, ionstage in ions:
        nnion = ionpopdict[(Z, ionstage)]
        if nnion == 0.:
            print(f'skipping Z={Z:2d} ion_stage {ionstage} due to nnion = {nnion}')
            continue
        X_ion = nnion / nntot
        dfcollion_thision = dfcollion.query('Z == @Z and ionstage == @ionstage', inplace=False)
        # if dfcollion.empty:
        #     continue
        ionpot_valence = dfcollion_thision.ionpot_ev.min()

        print(f'====> Z={Z:2d} ion_stage {ionstage} {at.get_ionstring(Z, ionstage)} (valence potential {ionpot_valence:.1f} eV)')

        print(f'               nnion: {nnion:.2e} [/cm3]')
        print(f'         nnion/nntot: {X_ion:.5f}')

        frac_ionization_ion[(Z, ionstage)] = 0.
        # integralgamma = 0.
        eta_over_ionpot_sum = 0.
        for index, shell in dfcollion_thision.iterrows():
            ar_xs_array = at.nonthermal.get_arxs_array_shell(engrid, shell)

            frac_ionization_shell = (
                nnion * shell.ionpot_ev * np.dot(yvec, ar_xs_array) * deltaen / deposition_density_ev)
            print(f'frac_ionization_shell(n {int(shell.n):d} l {int(shell.l):d}): '
                  f'{frac_ionization_shell:.4f} (ionpot {shell.ionpot_ev:.2f} eV)')

            # integralgamma += np.dot(yvec, ar_xs_array) * deltaen * shell.ionpot_ev / ionpot_valence

            if frac_ionization_shell > 1:
                frac_ionization_shell = 0.
                print(f'Ignoring frac_ionization_shell of {frac_ionization_shell}.')
                # for k in range(10):
                #     print(nnion * shell.ionpot_ev * yvec_reference[k] * ar_xs_array[k] * deltaen / E_init_ev)

            frac_ionization_ion[(Z, ionstage)] += frac_ionization_shell
            eta_over_ionpot_sum += frac_ionization_shell / shell.ionpot_ev

        frac_ionization += frac_ionization_ion[(Z, ionstage)]

        eff_ionpot_2 = X_ion / eta_over_ionpot_sum if eta_over_ionpot_sum else float('inf')

        eff_ionpot = (
            ionpot_valence * X_ion / frac_ionization_ion[(Z, ionstage)]
            if frac_ionization_ion[(Z, ionstage)] > 0. else float('inf'))

        print(f'     frac_ionization: {frac_ionization_ion[(Z, ionstage)]:.4f}')
        if not noexcitation:
            if nnion > 0.:
                frac_excitation_ion[(Z, ionstage)] = calculate_nt_frac_excitation(
                    engrid, dftransitions[(Z, ionstage)], yvec, deposition_density_ev)
            else:
                frac_excitation_ion[(Z, ionstage)] = 0.

            if frac_excitation_ion[(Z, ionstage)] > 1:
                frac_excitation_ion[(Z, ionstage)] = 0.
                print(f'Ignoring invalid frac_excitation_ion of {frac_excitation_ion[(Z, ionstage)]}.')
            frac_excitation += frac_excitation_ion[(Z, ionstage)]
            print(f'     frac_excitation: {frac_excitation_ion[(Z, ionstage)]:.4f}')
        else:
            frac_excitation_ion[(Z, ionstage)] = 0.

        print(f' eff_ionpot_shellpot: {eff_ionpot_2:.2f} [eV]')
        print(f'  eff_ionpot_valence: {eff_ionpot:.2f} [eV]')
        gamma_nt[(Z, ionstage)] = deposition_density_ev / nntot / eff_ionpot
        print(f'  Spencer-Fano Gamma: {gamma_nt[(Z, ionstage)]:.2e} [/s]')
        # print(f'Alternative Gamma: {integralgamma:.2e}')

        binding = get_mean_binding_energy(Z, ionstage, electron_binding, ionpot_ev=ionpot_valence)  # binding in erg
        oneoverW = Aconst * binding / Zbar / (2 * 3.14159 * pow(QE, 4))  # per erg
        deposition_density_erg = deposition_density_ev * EV

        # to get the non-thermal ionization rate we need to divide the energy deposited
        # per unit volume per unit time in the grid cell (sum of terms above)
        # by the total ion number density and the "work per ion pair"
        print(f"       work function: {1. / oneoverW / EV:.2f} [eV]")
        print(f"   work fn ratecoeff: {deposition_density_erg / nntot * oneoverW:.2e}")
        # Axelrod 1980 Eq 3.225 with E0 = E = E_max
        # xs = lossfunction(engrid[-1], nne, nnetot) * EV * oneoverW / nntot
        # print(f"         WFApprox xs: {xs:.2e} cm^2")

        print()

    # nne_nt = get_nne_nt(engrid, yvec)
    # print(f'               nne_nt: {nne_nt:.2e} /s/cm3')

    print(f'  frac_excitation_tot: {frac_excitation:.4f}')
    print(f'  frac_ionization_tot: {frac_ionization:.4f}')

    frac_heating = calculate_frac_heating(
        engrid, ions, ionpopdict, dfcollion, dftransitions, yvec, nne, nnetot, deposition_density_ev,
        noexcitation=noexcitation)

    print(f'         frac_heating: {frac_heating:.4f}')
    print(f'             frac_sum: {frac_excitation + frac_ionization + frac_heating:.4f}')

    return frac_excitation, frac_ionization, frac_heating, frac_excitation_ion, frac_ionization_ion, gamma_nt


def get_Latom_axelrod(Zboundbar, en_ev):
    # Axelrod 1980 Eq 3.21
    # Latom is 1/N * dE/dX where E is in erg
    # should be units of erg cm^2

    en_erg = en_ev * EV

    # relativistic
    gamma = en_erg / (ME * CLIGHT ** 2) + 1
    beta = math.sqrt(1. - 1. / (gamma ** 2))
    vel = beta * CLIGHT  # in cm/s

    # classical
    # vel = math.sqrt(2. * en_erg / ME)
    # beta = vel / CLIGHT

    # I = ionpot_ev * EV
    I = 280 * EV  # assumed in Axelrod thesis

    if 2 * ME * vel ** 2 < I:
        return 0.

    # if beta > 1.:
    #     print(vel, beta)
    #     beta = 0.9999

    return 4 * math.pi * QE ** 4 / (ME * vel ** 2) * Zboundbar * (
        math.log(2 * ME * vel ** 2 / I) + math.log(1. / (1. - beta ** 2)) - beta ** 2)


def get_Lelec_axelrod(en_ev, nne, nnetot, nntot):
    # - 1/N * dE / dX [erg cm^2]
    # returns a positive number

    # Axelrod Eq 3.36 (classical low energy limit)

    # return 1.95e-13 * math.log(3.2e4 * en_ev) / en_ev

    # Axelrod 1980 Eq 3.24

    HBAR = H / 2. / math.pi
    en_erg = en_ev * EV
    gamma = en_erg / (ME * CLIGHT ** 2) + 1
    beta = math.sqrt(1. - 1. / (gamma ** 2))
    vel = beta * CLIGHT  # in cm/s
    omegap = 5.64e4 * math.sqrt(nne)  # in per second
    return 4 * math.pi * QE ** 4 / (ME * vel ** 2) * nne / nntot * (
        math.log(2 * ME * vel ** 2 / (HBAR * omegap)) + 0.5 * math.log(1. / (1. - beta ** 2)) - 0.5 * beta ** 2)


def lossfunction_axelrod(en_ev, nne, nnetot):
    # - dE / dX [erg / cm]
    # returns a positive number

    return get_Lelec_axelrod(en_ev, nne=nne, nnetot=nnetot, nntot=1)


def get_fij_ln_en_ionisation(emax_ev, J, shell):
    npts = 1000
    e_p_lower = shell.ionpot_ev
    e_p_upper = emax_ev
    delta_e_p = (e_p_upper - e_p_lower) / npts
    sum = 0.
    for i in range(npts):
        e_p = e_p_lower + i * delta_e_p
        print(i, e_p)

        sigma = at.nonthermal.ar_xs(e_p, shell.ionpot_ev, shell.A, shell.B, shell.C, shell.D)
        eps_avg = get_epsilon_avg(e_p, J, shell.ionpot_ev)
        if eps_avg > 0:
            sum += ME * CLIGHT / math.pi / (QE ** 2) / H * sigma * math.log(eps_avg) * delta_e_p

    return sum


def e_s_test(ax, ionpot_ev, J, arr_en_ev, shellstr, color):
    e_p = arr_en_ev[-1]
    npts = 100000
    # ionpot_ev = 280

    J = ionpot_ev * 0.6
    # epsilon_lower = ionpot_ev
    # epsilon_upper = (ionpot_ev + e_p) / 2.
    # delta_eps = (epsilon_upper - epsilon_lower) / npts
    # print(f'{delta_eps=}')
    # prob_sum = 0.
    # eps_avg = 0.
    # prob = np.zeros(npts)
    # epsilon = np.zeros(npts)
    # for i in range(npts):
    #     epsilon[i] = epsilon_lower + i * delta_eps
    #     prob[i] = Psecondary(e_p, ionpot_ev, J, epsilon=epsilon[i]) * delta_eps
    #     prob_sum += prob[i]
    #     eps_avg = epsilon[i] * prob[i]
    # print(f'{prob_sum=:.1e}')
    # print(f'{eps_avg=:.1e}')
    # # ax.plot(epsilon, prob, label=shellstr + ' dP / d_epsilon')

    e_s_lower = 0
    e_s_upper = (e_p - ionpot_ev) / 2.
    delta_e_s = (e_s_upper - e_s_lower) / npts
    prob_sum = 0.
    e_s_avg = 0.
    prob = np.zeros(npts)
    e_s = np.zeros(npts)
    prob_e_s_can_ionise = 0.
    prob_e_s_cannot_ionise = 0.
    for i in range(npts):
        e_s[i] = e_s_lower + i * delta_e_s
        prob[i] = Psecondary(e_p, ionpot_ev, J, e_s=e_s[i]) * delta_e_s
        prob_sum += prob[i]
        e_s_avg = e_s[i] * prob[i]
        if e_s[i] >= ionpot_ev:
            prob_e_s_can_ionise += prob[i]
        else:
            prob_e_s_cannot_ionise += prob[i]

    print(f'prob_sum={prob_sum:.2f}')
    print(f'e_s_avg={e_s_avg:.1e}')
    print(f'prob_e_s_can_ionise={prob_e_s_can_ionise:.2f}')
    print(f'prob_e_s_cannot_ionise={prob_e_s_cannot_ionise:.2f}')
    # ax.plot(e_s, prob, label=shellstr + ' dP / d_e_s', color=color)
    # ax.vlines(ionpot_ev, ymin=0., ymax=max(prob), color=color)


def get_epsilon_avg(e_p, J, ionpot_ev, quiet=True):
    # average energy loss of the primary electron per ionisation in eV
    npts = 1000000

    epsilon_lower = ionpot_ev
    epsilon_upper = (ionpot_ev + e_p) / 2.
    if epsilon_upper <= epsilon_lower:
        return 0.
    delta_eps = (epsilon_upper - epsilon_lower) / npts
    # print(f'{delta_eps=}')
    # print(f'{epsilon_lower} {epsilon_upper} {delta_eps} {e_p}')
    prob_sum = 0.
    prob_e_s_can_ionise = 0.
    prob_e_s_cannot_ionise = 0.
    eps_avg = 0.
    e_s_avg = 0.
    for i in range(npts):
        epsilon = epsilon_lower + i * delta_eps
        prob = Psecondary(e_p, ionpot_ev, J, epsilon=epsilon) * delta_eps
        prob_sum += prob
        eps_avg += epsilon * prob
        e_s = epsilon - ionpot_ev
        e_s_avg += e_s * prob
        if e_s >= ionpot_ev:
            prob_e_s_can_ionise += prob
        else:
            prob_e_s_cannot_ionise += prob
    # print(f'{prob_sum=:.3f}')
    assert(abs(prob_sum - 1.) < 0.30)
    if not quiet:
        print(f'e_p: {e_p:.1e}')
        print(f'eps_avg: {eps_avg:.1e}')
        print(f'e_s_avg: {e_s_avg:.1e}')
        print(f'prob_e_s_can_ionise: {prob_e_s_can_ionise:.2f}')
        print(f'prob_e_s_cannot_ionise: {prob_e_s_cannot_ionise:.2f}')
    return eps_avg


def calculate_Latom_excitation(ions, ionpopdict, nntot, en_ev, adata, T_exc=5000):
    L_atom_sum = 0.
    for Z, ionstage in ions:
        nnion = ionpopdict[(Z, ionstage)]

        ion = adata.query('Z == @Z and ion_stage == @ionstage').iloc[0]
        dftransitions_ion = ion.transitions

        maxnlevelslower = 5
        maxnlevelsupper = 250

        # filterquery = 'collstr >= 0 or forbidden == False'
        filterquery = 'collstr != -999'
        if maxnlevelslower is not None:
            filterquery += ' and lower < @maxnlevelslower'
        if maxnlevelsupper is not None:
            filterquery += ' and upper < @maxnlevelsupper'

        dftransitions_ion = dftransitions_ion.query(
            filterquery, inplace=False).copy()

        k_b = 8.617333262145179e-05  # eV / K
        energy_boltzfac_sum = ion.levels.eval(
            'energy_ev * g * exp(- energy_ev / @k_b / @T_exc)').sum()

        populations = ion.levels.eval('g * exp(- energy_ev / @k_b / @T_exc)').values / energy_boltzfac_sum

        dftransitions_ion.eval(
            'epsilon_trans_ev = '
            '@ion.levels.loc[upper].energy_ev.values - @ion.levels.loc[lower].energy_ev.values', inplace=True)

        dftransitions_ion.eval('upper_g = @ion.levels.loc[upper].g.values', inplace=True)
        dftransitions_ion.eval('lower_g = @ion.levels.loc[lower].g.values', inplace=True)

        for _, row in dftransitions_ion.iterrows():
            nnlevel = populations[row.lower] * nnion
            epsilon_trans_ev = row.epsilon_trans_ev
            sigma = get_xs_excitation(en_ev, row)
            L_atom_sum += sigma * epsilon_trans_ev * EV * nnlevel / nntot

    return L_atom_sum


def calculate_Latom_ionisation(ions, ionpopdict, dfcollion, nntot, nne, en_ev, nnetot=None, electron_binding=None):
    L_atom_sum = 0.
    for Z, ionstage in ions:
        nnion = ionpopdict[(Z, ionstage)]
        dfcollion_thision = dfcollion.query('Z == @Z and ionstage == @ionstage', inplace=False)
        # ionpot_valence_ev = dfcollion_thision.ionpot_ev.min()

        for _, shell in dfcollion_thision.iterrows():
            J = get_J(Z, ionstage, shell.ionpot_ev)

            eps_avg = get_epsilon_avg(en_ev, J, shell.ionpot_ev)
            # eps_avg = shell.ionpot_ev
            if eps_avg > 0:
                # sigma = get_lotz_xs_ionisation(Z, ionstage, electron_binding, ionpot_ev=ionpot_valence_ev, en_ev=en_ev)
                sigma = at.nonthermal.ar_xs(en_ev, shell.ionpot_ev, shell.A, shell.B, shell.C, shell.D)
                # L_atom_sum += ME * CLIGHT / math.pi / (QE ** 2) * sigma * eps_avg * nnion / nntot
                L_atom_sum += sigma * eps_avg * EV * nnion / nntot

        # J = get_J(Z, ionstage, ionpot_valence_ev)
        # eps_avg = get_epsilon_avg(en_ev, J, ionpot_valence_ev)
        # # test case: secondary electron has no energy
        # # eps_avg = ionpot_valence_ev
        # if eps_avg > 0:
        #     sigma = get_lotz_xs_ionisation(Z, ionstage, electron_binding, ionpot_ev=ionpot_valence_ev, en_ev=en_ev)
        #     # sigma = at.nonthermal.ar_xs(en_ev, shell.ionpot_ev, shell.A, shell.B, shell.C, shell.D)
        #     # L_atom_sum += ME * CLIGHT / math.pi / (QE ** 2) * sigma * eps_avg * nnion / nntot
        #     # nnebound = nnetot - nne
        #     # 1e-4 = EV * 2 * 3.14 * 1e7
        #     L_atom_sum += sigma * eps_avg * EV * nnion / nntot

        # print(f'ln I: {ln_I_sum:.2e} (after ionsiation)')
        # print(f'I: {math.exp(ln_I_sum):.2e} (after ionisation)')

    return L_atom_sum


def workfunction_tests(modelpath, args):
    electron_binding = read_binding_energies()
    dfcollion = at.nonthermal.read_colliondata()

    fig, axes = plt.subplots(nrows=1, ncols=1, sharex=True,
                             figsize=(6*.7, 5*.7), tight_layout={"pad": 0.3, "w_pad": 0.0, "h_pad": 0.0})
    axes = [axes]

    ionpopdict = {
        # (16, 2): 3e5,
        (26, 2): 1e5,
    }

    ions = []
    for key in ionpopdict.keys():
        # keep only the ion populations, not element or total populations
        if isinstance(key, tuple) and len(key) == 2:
            ions.append(key)

    ions.sort()

    nntot = get_nntot(ions=ions, ionpopdict=ionpopdict)
    nnetot = get_nnetot(ions=ions, ionpopdict=ionpopdict)  # total electrons: free and bound included
    nne = get_nne(ions=ions, ionpopdict=ionpopdict)
    nnebound = nnetot - nne
    # x_e = nne / nntot
    Zbar = get_Zbar(ions=ions, ionpopdict=ionpopdict)  # average atomic number
    Zboundbar = get_Zboundbar(ions=ions, ionpopdict=ionpopdict)  # average bound electrons per nucleus
    print(f'nntot: {nntot:.1e}')
    print(f'nnetot:{nnetot:.1e}')
    print(f'nne: {nne:.1e}')
    print(f'nnebound: {nnebound:.1e}')
    print(f'Zbar: {Zbar:.1e}')
    print(f'Zboundbar: {Zboundbar:.1e}')

    hbar_ev_s = 6.58211951e-16  # in eV seconds
    omegap = 5.6e4 * math.sqrt(nne)  # in per second
    print(f'hbar * omegap = {hbar_ev_s * omegap:.2e} eV')

    en_min_ev = args.emin
    en_max_ev = args.emax
    print(f'en_min_ev: {en_min_ev:.1f} eV')
    print(f'en_max_ev: {en_max_ev:.1f} eV')
    # arr_en_ev = np.linspace(en_min_ev, en_max_ev, 2000)
    # delta_en_ev = arr_en_ev[1] - arr_en_ev[0]
    arr_en_ev = np.logspace(
        start=math.log10(en_min_ev), stop=math.log10(en_max_ev), base=10, num=args.npts, endpoint=True
    )

    global Psecondary_e_s_max
    Psecondary_e_s_max = arr_en_ev[2]
    print(f'Psecondary_e_s_max: {Psecondary_e_s_max}')

    delta_en_ev = arr_en_ev[1:] - arr_en_ev[:-1]
    arr_en_ev = arr_en_ev[:-1]  # remove the endpoint, now that we've used it to calculate detla_en_ev

    # Axelrod 1980 Eq 3.24
    Lelec_axelrod_nne = np.array([get_Lelec_axelrod(en_ev=en_ev, nne=nne, nnetot=nnetot, nntot=nntot) for en_ev in arr_en_ev])
    # Lelec_axelrod_nnetot = np.array([get_Lelec_axelrod(en_ev=en_ev, nne=nnetot, nnetot=nnetot, nntot=nntot) for en_ev in arr_en_ev])

    Lelec_kf92_nne = np.array([lossfunction(en_ev, nne, nnetot, use_nnetot=False) * EV / nntot for en_ev in arr_en_ev])

    Lelec_kf92_nnetot = np.array([lossfunction(en_ev, nne, nnetot, use_nnetot=True) * EV / nntot for en_ev in arr_en_ev])

    Lelec_kf92_nnebound = np.array([lossfunction(en_ev, nnebound, nnetot, use_nnetot=False) * EV / nntot for en_ev in arr_en_ev])

    # print(f'{Lelec_axelrod[-1]=}')
    # print(f'{Lelec_kf92[-1]=}')

    Latom_axelrod = np.array([get_Latom_axelrod(en_ev=en_ev, Zboundbar=Zboundbar) for en_ev in arr_en_ev])

    # adata = at.atomic.get_levels(modelpath, get_transitions=False, ionlist=tuple(ions))
    # adata = None
    arr_Latom_summed = np.zeros_like(arr_en_ev)
    for i, en_ev in enumerate(arr_en_ev):
        arr_Latom_summed[i] = calculate_Latom_ionisation(
            ions=ions, ionpopdict=ionpopdict, dfcollion=dfcollion, nntot=nntot, nne=nne,
            en_ev=en_ev, nnetot=nnetot, electron_binding=electron_binding)

        print(f'{en_ev:.2f} eV L_atom_summed: {arr_Latom_summed[i]:.3e} (ionisation only) '
              f'Latom_axelrod: {Latom_axelrod[i]:.3e} ratio(sum/axelrod): {arr_Latom_summed[i] / Latom_axelrod[i]:.2e} ratio(axelrod/sum) {Latom_axelrod[i] / arr_Latom_summed[i]:.2e} [erg cm2]')
        # print(f'{en_ev:.2f} eV L_elec_axelrod: {Lelec_axelrod_nne[i]:.3e} '
        #       f'Latom_axelrod: {Latom_axelrod[i]:.3e} ratio(elec/atom): {Lelec_axelrod_nne[i] / Latom_axelrod[i]:.2e} ratio(atom/elec) {Latom_axelrod[i] / Lelec_axelrod_nne[i]:.2e} bound/free {nnebound/nne:.2e}')

    for Z, ionstage in ions:
        # ionstr = at.get_ionstring(Z, ionstage, spectral=True, nospace=False)
        dfcollion_thision = dfcollion.query('Z == @Z and ionstage == @ionstage', inplace=False)

        # dfcollion_thision.query('n == 3 and l == 2', inplace=True)
        # dfcollion_thision.eval('A = 0', inplace=True)
        # dfcollion_thision.eval('B = 0', inplace=True)
        # # dfcollion_thision.eval('C = 0', inplace=True)
        # dfcollion_thision.eval('D = 0', inplace=True)

        # shellindex = 0
        # for _, shell in dfcollion_thision.iterrows():
        #     shellstr = f'\nn {shell.n} l {shell.l} ionpot {shell.ionpot_ev} eV'
        #     print(shellstr)
        #     e_p = arr_en_ev[-1]
        #     J = get_J(Z, ionstage, shell.ionpot_ev)
        #     e_s_test(axes[-1], shell.ionpot_ev, J, arr_en_ev, shellstr, color=f'C{shellindex}')
        #     shellindex += 1

        ionpot_valence_ev = dfcollion_thision.ionpot_ev.min()
        print(f'\n===> ion Z {Z} ionstage {ionstage} ionpot_valence_ev {ionpot_valence_ev}')
        # print(dfcollion_thision)

        binding_alt = get_mean_binding_energy_alt(Z, ionstage, electron_binding, ionpot_ev=ionpot_valence_ev)  # binding in erg
        print(f'mean binding energy: {binding_alt / EV:.2f} eV')
        Aconst = 1.33e-14 * EV * EV
        binding = get_mean_binding_energy(Z, ionstage, electron_binding, ionpot_ev=ionpot_valence_ev)  # binding in erg
        oneoverW_limit_sim = Aconst * binding / Zbar / (2 * 3.14159 * pow(QE, 4))  # per erg
        workfn_limit_ev_sim = 1. / oneoverW_limit_sim / EV
        print(f'\n workfn_limit_ev_sim {workfn_limit_ev_sim:.2f} eV')
        print(f'   eta_ion  {ionpot_valence_ev / workfn_limit_ev_sim:.3f}')
        print(f'   eta_heat {1 - ionpot_valence_ev / workfn_limit_ev_sim:.3f}')
        arr_workfn_limit_sim = np.array([workfn_limit_ev_sim for x in arr_en_ev])

        arr_xs_ar92 = at.nonthermal.get_arxs_array_ion(arr_en_ev, dfcollion_thision, Z, ionstage)
        Latom_ionisation_ar92 = arr_xs_ar92 * (ionpot_valence_ev * EV)

        arr_xs_lotz = np.array([
            get_lotz_xs_ionisation(Z, ionstage, electron_binding, ionpot_ev=ionpot_valence_ev, en_ev=en_ev)
            for en_ev in arr_en_ev])

        axes[-1].plot(arr_en_ev, Latom_axelrod, label='Latom_axelrod')

        # Axelrod 1980 Eq 3.20, (Latom part).
        # This assumes that the every bound electron cross section is included!
        Latom_ionisation_lotz = arr_xs_lotz * (ionpot_valence_ev * EV)

        L_over_sigma = (Lelec_axelrod_nne[-1] + Latom_axelrod[-1]) / arr_xs_lotz[-1]
        workfn_limit_axelrod = L_over_sigma / EV
        print(f'\n workfn_limit_axelrod: {workfn_limit_axelrod:.2f} eV')
        print(f'   eta_ion  {ionpot_valence_ev / workfn_limit_axelrod:.3f}')
        print(f'   eta_heat {1 - ionpot_valence_ev / workfn_limit_axelrod:.3f}')
        arr_workfn_limit_axelrod = np.array([workfn_limit_axelrod for x in arr_en_ev])

        # Approximation to Axelrod 1980 Eq 3.20 (left Latom part) where the transition energy
        # of every ionisation is just the valence potential
        # Latom_ionisation = arr_xs * (ionpot_valence_ev * EV)
        # print(Latom[-1], Latom2[-1], Latom3[-1])

        # arr_xs_latom = Latom1 / (ionpot_valence_ev * EV)

        # axes[-1].plot(arr_en_ev, arr_xs_lotz, label=r'$\sigma_{Lotz}$')
        # axes[-1].plot(arr_en_ev, arr_xs_ar92, label=r'$\sigma_{AR92}$')
        # axes[-1].plot(arr_en_ev, arr_xs_latom, label=r'$\sigma$=$L_{atom}/I$', linestyle='dashed')

        axes[-1].plot(arr_en_ev, arr_Latom_summed, label=r'Latom_summed')

        # Lelec = Lelec_kf92_nne
        # Lelec = Lelec_kf92_nnetot
        # Lelec = Lelec_kf92_nnebound
        Lelec = Lelec_axelrod_nne
        # Lelec = Lelec_axelrod_nnetot

        Latom = Latom_axelrod
        # Latom = Latom_ionisation_lotz
        # Latom = Latom_ionisation_ar92
        # axes[-1].plot(arr_en_ev, Lelec_kf92_nne / Latom_axelrod, label='Lelec_kf92 / Latom_axelrod')
        # axes[-1].plot(arr_en_ev, Lelec_axelrod / Latom_axelrod, label='Lelec_axelrod / Latom_axelrod')

        # arr_Lelec_over_Latom = [x_e / (Zbar - x_e) * (math.log(4 * en_ev / (hbar_ev_s * omegap)) / math.log(4 * en_ev / 300)) for en_ev in arr_en_ev]
        # axes[-1].plot(arr_en_ev, arr_Lelec_over_Latom, label='Lelec_axelrod / Latom_axelrod analytic')

        # axes[-1].plot(arr_en_ev, Lelec_axelrod / Lelec_kf92_nne, label='lossfunction_axelrod / lossfunction_kf92')

        # Latom = Latom_axelrod - Latom_ionisation_lotz + Latom_ionisation_ar92

        # arr_xs = arr_xs_lotz
        arr_xs = arr_xs_ar92
        # axes[-1].plot(arr_en_ev, arr_xs_ar92 * 1e18, label=ionstr + ' Arnaud & Rotherflug 1992', linewidth=2)
        # axes[-1].plot(arr_en_ev, arr_xs_lotz * 1e18, label=ionstr + ' Lotz (1967)', linewidth=2)

        L = Lelec + Latom

        with np.errstate(divide='ignore'):
            workfn_limit = L / EV / arr_xs

        print(f'\n workfn_limit at Emax: {workfn_limit[-1]:.2f} eV')
        print(f'   eta_ion  {ionpot_valence_ev / workfn_limit[-1]:.3f}')
        print(f'   eta_heat {1 - ionpot_valence_ev / workfn_limit[-1]:.3f}')

        arr_workfn_integrated = np.zeros_like(arr_en_ev)
        integrand = arr_xs / (L / EV)
        # arr_workfn_integrated[i] is the en_ev / (integral xs / L dE from EMIN to E[i])
        with np.errstate(divide='ignore'):
            arr_workfn_integrated = [arr_en_ev[i] / (sum((integrand * delta_en_ev)[:i])) for i in range(len(arr_en_ev))]

        print(f'\n workfn_integral_Emin_Emax: {arr_workfn_integrated[-1]:.2f} eV')
        print(f'   eta_ion  {ionpot_valence_ev / arr_workfn_integrated[-1]:.3f}')
        print(f'   eta_heat {1 - ionpot_valence_ev / arr_workfn_integrated[-1]:.3f}')

        # axes[-1].plot(arr_en_ev, arr_workfn_limit_axelrod, label='workfn limit E->inf (Axelrod)')
        # axes[-1].plot(arr_en_ev, arr_workfn_limit_sim, label='workfn limit E->inf (Sim)')
        # axes[-1].plot(arr_en_ev, workfn_limit, label='workfn integrated near Emax', color='C2')
        # axes[-1].plot(arr_en_ev, arr_workfn_integrated, label='workfn integrated Emin to Emax', color='C3')

        # arr = [lossfunction(en_ev, nne, nnetot) / arr_xs_ar92[i] for i, en_ev in enumerate(arr_en_ev)]
        # arr = [Latom_axelrod[i] / arr_xs_ar92[i] for i, en_ev in enumerate(arr_en_ev)]

        # print(arr)
        # print(min(arr), max(arr))
        # axes[-1].plot(arr_en_ev, arr, label='lossfunc / sigma', color='C3')

    axes[-1].set_xlabel(r'Energy [eV]')
    axes[-1].set_xscale('log')
    # axes[-1].set_xlim(0., 500)
    # axes[-1].set_ylim(0., 2000)
    # axes[-1].set_ylim(bottom=1e-3, top=1e-2)
    # axes[-1].set_ylim(bottom=min(arr) / 2., top=max(arr) * 2.)
    # axes[-1].set_ylabel(r'cross section [cm$^{2}$]')
    # axes[-1].set_ylabel(r'Cross Section $[10^{-18}$ cm$^{2}]$')
    # axes[-1].set_ylabel(r'Ionisation fraction of deposited energy')
    axes[-1].set_yscale('log')
    # axes[-1].set_ylabel(r'log y(E) [s$^{-1}$ cm$^{-2}$ eV$^{-1}$]', fontsize=fs)
    # axes[-1].yaxis.set_minor_locator(ticker.MultipleLocator(base=5))
    axes[-1].legend(frameon=False, loc='upper right')
    outputfilename = 'plot.pdf'
    print(f"Saving '{outputfilename}'")
    fig.savefig(str(outputfilename), format='pdf')
    plt.close()



#!/usr/bin/env python3
"""Artistools - spectra related functions."""
import argparse
import json
import math
import multiprocessing
from collections import namedtuple
# from functools import lru_cache
from functools import partial
from pathlib import Path

import matplotlib as mpl
import matplotlib.pyplot as plt
# import matplotlib.ticker as ticker
import numpy as np
import pandas as pd
# from astropy import constants as const
from astropy import units as u

import artistools as at
import artistools.packets


def get_packets_with_emtype_onefile(emtypecolumn, lineindices, packetsfile):
    import gzip
    try:
        dfpackets = at.packets.readfile(packetsfile, type='TYPE_ESCAPE', escape_type='TYPE_RPKT')
    except gzip.BadGzipFile:
        print(f'Bad file: {packetsfile}')
        raise gzip.BadGzipFile

    return dfpackets.query(f'{emtypecolumn} in @lineindices', inplace=False).copy()


@at.diskcache(savezipped=True)
def get_packets_with_emtype(modelpath, emtypecolumn, lineindices, maxpacketfiles=None):
    packetsfiles = at.packets.get_packetsfilepaths(modelpath, maxpacketfiles=maxpacketfiles)
    nprocs_read = len(packetsfiles)
    assert nprocs_read > 0

    model, _, _ = at.inputmodel.get_modeldata(modelpath)
    # vmax = model.iloc[-1].velocity_outer * u.km / u.s
    processfile = partial(get_packets_with_emtype_onefile, emtypecolumn, lineindices)
    if at.num_processes > 1:
        print(f"Reading packets files with {at.num_processes} processes")
        with multiprocessing.Pool(processes=at.num_processes) as pool:
            arr_dfmatchingpackets = pool.map(processfile, packetsfiles)
            pool.close()
            pool.join()
            # pool.terminate()
    else:
        arr_dfmatchingpackets = [processfile(f) for f in packetsfiles]

    dfmatchingpackets = pd.concat(arr_dfmatchingpackets)

    return dfmatchingpackets, nprocs_read


def calculate_timebinned_packet_sum(dfpackets, timearrayplusend):
    binned = pd.cut(dfpackets['t_arrive_d'], timearrayplusend, labels=False, include_lowest=True)

    binnedenergysums = np.zeros_like(timearrayplusend[:-1], dtype=float)
    for binindex, e_rf_sum in dfpackets.groupby(binned)['e_rf'].sum().iteritems():
        binnedenergysums[int(binindex)] = e_rf_sum

    return binnedenergysums


def get_line_fluxes_from_packets(emtypecolumn, emfeatures, modelpath, maxpacketfiles=None, arr_tstart=None, arr_tend=None):
    if arr_tstart is None:
        arr_tstart = at.get_timestep_times_float(modelpath, loc='start')
    if arr_tend is None:
        arr_tend = at.get_timestep_times_float(modelpath, loc='end')

    arr_timedelta = np.array(arr_tend) - np.array(arr_tstart)
    arr_tmid = arr_tend = (np.array(arr_tstart) + np.array(arr_tend)) / 2.

    model, _, _ = at.inputmodel.get_modeldata(modelpath)
    # vmax = model.iloc[-1].velocity_outer * u.km / u.s
    # betafactor = math.sqrt(1 - (vmax / const.c).decompose().value ** 2)

    timearrayplusend = np.concatenate([arr_tstart, [arr_tend[-1]]])

    dictlcdata = {'time': arr_tmid}

    linelistindices_allfeatures = tuple([l for feature in emfeatures for l in feature.linelistindices])

    dfpackets, nprocs_read = get_packets_with_emtype(
            modelpath, emtypecolumn, linelistindices_allfeatures, maxpacketfiles=maxpacketfiles)

    for feature in emfeatures:
        # dictlcdata[feature.colname] = np.zeros_like(arr_tstart, dtype=float)

        dfpackets_selected = dfpackets.query(f'{emtypecolumn} in @feature.linelistindices', inplace=False)

        normfactor = 1. / nprocs_read
        # mpc_to_cm = 3.085677581491367e+24  # 1 megaparsec in cm
        # normfactor = 1. / 4 / math.pi / (mpc_to_cm ** 2) / nprocs_read

        energysumsreduced = calculate_timebinned_packet_sum(dfpackets_selected, timearrayplusend)
        # print(energysumsreduced, arr_timedelta)
        fluxdata = np.divide(energysumsreduced * normfactor, arr_timedelta * u.day.to('s'))
        dictlcdata[feature.colname] = fluxdata

    lcdata = pd.DataFrame(dictlcdata)
    return lcdata


def get_line_fluxes_from_pops(emtypecolumn, emfeatures, modelpath, arr_tstart=None, arr_tend=None):
    import artistools.nltepops
    if arr_tstart is None:
        arr_tstart = at.get_timestep_times_float(modelpath, loc='start')
    if arr_tend is None:
        arr_tend = at.get_timestep_times_float(modelpath, loc='end')

    arr_timedelta = np.array(arr_tend) - np.array(arr_tstart)
    arr_tmid = arr_tend = (np.array(arr_tstart) + np.array(arr_tend)) / 2.

    modeldata, _, _ = at.inputmodel.get_modeldata(modelpath)

    ionlist = []
    for feature in emfeatures:
        ionlist.append((feature.atomic_number, feature.ion_stage))

    adata = at.atomic.get_levels(modelpath, ionlist=tuple(ionlist), get_transitions=True, get_photoionisations=False)

    timearrayplusend = np.concatenate([arr_tstart, [arr_tend[-1]]])

    dictlcdata = {'time': arr_tmid}

    for feature in emfeatures:
        fluxdata = np.zeros_like(arr_tmid, dtype=float)

        dfnltepops = at.nltepops.read_files(
            modelpath,
            dfquery=f'Z=={feature.atomic_number:.0f} and ion_stage=={feature.ion_stage:.0f}').query('level in @feature.upperlevelindicies')

        ion = adata.query(
            'Z == @feature.atomic_number and ion_stage == @feature.ion_stage').iloc[0]

        for timeindex, timedays in enumerate(arr_tmid):
            v_inner = modeldata.velocity_inner.values * u.km / u.s
            v_outer = modeldata.velocity_outer.values * u.km / u.s

            t_sec = timedays * u.day
            shell_volumes = ((4 * math.pi / 3) * ((v_outer * t_sec) ** 3 - (v_inner * t_sec) ** 3)).to('cm3').value

            timestep = at.get_timestep_of_timedays(modelpath, timedays)
            print(f'{feature.approxlambda}A {timedays}d (ts {timestep})')

            for upperlevelindex, lowerlevelindex in zip(feature.upperlevelindicies, feature.lowerlevelindicies):
                unaccounted_shellvol = 0.  # account for the volume of empty shells
                unaccounted_shells = []
                for modelgridindex in modeldata.index:
                    try:
                        levelpop = dfnltepops.query(
                            'modelgridindex==@modelgridindex and timestep==@timestep and Z==@feature.atomic_number'
                            ' and ion_stage==@feature.ion_stage and level==@upperlevelindex').iloc[0].n_NLTE

                        A_val = ion.transitions.query(
                                'upper == @upperlevelindex and lower == @lowerlevelindex').iloc[0].A

                        delta_ergs = (
                            ion.levels.iloc[upperlevelindex].energy_ev -
                            ion.levels.iloc[lowerlevelindex].energy_ev) * u.eV.to('erg')

                        # l = delta_ergs * A_val * levelpop * (shell_volumes[modelgridindex] + unaccounted_shellvol)
                        # print(f'  {modelgridindex} outer_velocity {modeldata.velocity_outer.values[modelgridindex]}'
                        #       f' km/s shell_vol: {shell_volumes[modelgridindex] + unaccounted_shellvol} cm3'
                        #       f' n_level {levelpop} cm-3 shell_Lum {l} erg/s')

                        fluxdata[timeindex] += delta_ergs * A_val * levelpop * (
                            shell_volumes[modelgridindex] + unaccounted_shellvol)

                        unaccounted_shellvol = 0.

                    except IndexError:
                        unaccounted_shellvol += shell_volumes[modelgridindex]
                        unaccounted_shells.append(modelgridindex)
                if unaccounted_shells:
                    print(f'No data for cells {unaccounted_shells} (expected for empty cells)')
                assert len(unaccounted_shells) < len(modeldata.index)  # must be data for at least one shell

        dictlcdata[feature.colname] = fluxdata

    lcdata = pd.DataFrame(dictlcdata)
    return lcdata


def get_closelines(modelpath, atomic_number, ion_stage, approxlambda, lambdamin=-1, lambdamax=-1, lowerlevelindex=-1, upperlevelindex=-1):
    dflinelist = at.get_linelist(modelpath, returntype='dataframe')
    dflinelistclosematches = dflinelist.query('atomic_number == @atomic_number and ionstage == @ion_stage').copy()
    if lambdamin > 0:
        dflinelistclosematches.query('@lambdamin < lambda_angstroms', inplace=True)
    if lambdamax > 0:
        dflinelistclosematches.query('@lambdamax > lambda_angstroms', inplace=True)
    if lowerlevelindex >= 0:
        dflinelistclosematches.query('lowerlevelindex==@lowerlevelindex', inplace=True)
    if upperlevelindex >= 0:
        dflinelistclosematches.query('upperlevelindex==@upperlevelindex', inplace=True)
    # print(dflinelistclosematches)

    linelistindices = tuple(dflinelistclosematches.index.values)
    upperlevelindicies = tuple(dflinelistclosematches.upperlevelindex.values)
    lowerlevelindicies = tuple(dflinelistclosematches.lowerlevelindex.values)
    lowestlambda = dflinelistclosematches.lambda_angstroms.min()
    highestlamba = dflinelistclosematches.lambda_angstroms.max()
    colname = f'flux_{at.get_ionstring(atomic_number, ion_stage, nospace=True)}_{approxlambda}'
    featurelabel = f'{at.get_ionstring(atomic_number, ion_stage)} {approxlambda} Å'

    return (colname, featurelabel, approxlambda, linelistindices, lowestlambda, highestlamba,
            atomic_number, ion_stage, upperlevelindicies, lowerlevelindicies)


def get_labelandlineindices(modelpath, emfeaturesearch):
    featuretuple = namedtuple('feature', [
        'colname', 'featurelabel', 'approxlambda', 'linelistindices', 'lowestlambda',
        'highestlamba', 'atomic_number', 'ion_stage', 'upperlevelindicies', 'lowerlevelindicies'])

    labelandlineindices = []
    for params in emfeaturesearch:
        feature = featuretuple(*get_closelines(modelpath, *params))
        print(f'{feature.featurelabel} includes {len(feature.linelistindices)} lines '
              f'[{feature.lowestlambda:.1f} Å, {feature.highestlamba:.1f} Å]')
        labelandlineindices.append(feature)
    # labelandlineindices.append(featuretuple(*get_closelines(dflinelist, 26, 2, 7155, 7150, 7160)))
    # labelandlineindices.append(featuretuple(*get_closelines(dflinelist, 26, 2, 12570, 12470, 12670)))
    # labelandlineindices.append(featuretuple(*get_closelines(dflinelist, 28, 2, 7378, 7373, 7383)))

    return labelandlineindices


def make_flux_ratio_plot(args):
    # font = {'size': 16}
    # matplotlib.rc('font', **font)
    nrows = 1
    fig, axes = plt.subplots(
        nrows=nrows, ncols=1, sharey=False,
        figsize=(args.figscale * at.figwidth, args.figscale * at.figwidth * (0.25 + nrows * 0.4)),
        tight_layout={"pad": 0.2, "w_pad": 0.0, "h_pad": 0.0})

    if nrows == 1:
        axes = [axes]

    axis = axes[0]
    axis.set_yscale('log')
    # axis.set_ylabel(r'log$_1$$_0$ F$_\lambda$ at 1 Mpc [erg/s/cm$^2$/$\mathrm{{\AA}}$]')

    # axis.set_xlim(left=supxmin, right=supxmax)
    pd.set_option('display.max_rows', 3500)
    pd.set_option('display.width', 150)
    pd.options.display.max_rows = 500

    for seriesindex, (modelpath, modellabel, modelcolor) in enumerate(zip(args.modelpath, args.label, args.color)):
        print(f"====> {modellabel}")

        emfeatures = get_labelandlineindices(modelpath, tuple(args.emfeaturesearch))

        if args.frompops:
            dflcdata = get_line_fluxes_from_pops(args.emtypecolumn, emfeatures, modelpath,
                                                 arr_tstart=args.timebins_tstart,
                                                 arr_tend=args.timebins_tend)
        else:
            dflcdata = get_line_fluxes_from_packets(args.emtypecolumn, emfeatures, modelpath,
                                                    maxpacketfiles=args.maxpacketfiles,
                                                    arr_tstart=args.timebins_tstart,
                                                    arr_tend=args.timebins_tend)

        dflcdata.eval(f'fratio = {emfeatures[1].colname} / {emfeatures[0].colname}', inplace=True)
        axis.set_ylabel(r'F$_{\mathrm{' + emfeatures[1].featurelabel + r'}}$ / F$_{\mathrm{' +
                        emfeatures[0].featurelabel + r'}}$')

        # \mathrm{\AA}

        print(dflcdata)

        axis.plot(dflcdata.time, dflcdata['fratio'], label=modellabel, marker='x', lw=0,
                  markersize=10, markeredgewidth=2,
                  color=modelcolor, alpha=0.8, fillstyle='none')

        tmin = dflcdata.time.min()
        tmax = dflcdata.time.max()

    if args.emfeaturesearch[0][:3] == (26, 2, 7155) and args.emfeaturesearch[1][:3] == (26, 2, 12570):
        axis.set_ylim(ymin=0.05)
        axis.set_ylim(ymax=7)
        arr_tdays = np.linspace(tmin, tmax, 3)
        arr_floersfit = [10 ** (0.0043 * timedays - 1.65) for timedays in arr_tdays]
        for ax in axes:
            ax.plot(arr_tdays, arr_floersfit, color='black', label='Flörs+2020 fit', lw=2.)

        femis = pd.read_csv(
            "/Users/luke/Dropbox/Papers (first-author)/2022 Artis ionisation/"
            "generateplots/floers_model_NIR_VIS_ratio_20201126.csv")

        amodels = {}
        for index, row in femis.iterrows():
            modelname = row.file.replace("fig-nne_Te_allcells-", "").replace(f"-{row.epoch}d.txt", "")
            if modelname not in amodels:
                amodels[modelname] = ([], [])
            if int(row.epoch) != 263:
                amodels[modelname][0].append(row.epoch)
                amodels[modelname][1].append(row.NIR_VIS_ratio)

        aindex = 0
        # for amodelname, (xlist, ylist) in amodels.items():
        for amodelname, alabel in [
                # ('w7', 'W7'),
                # ('subch', 'S0'),
                # ('subch_shen2018', r'1M$_\odot$'),
                # ('subch_shen2018_electronlossboost4x', '1M$_\odot$ (Shen+18) 4x e- loss'),
                # ('subch_shen2018_electronlossboost8x', r'1M$_\odot$ heatboost8'),
                # ('subch_shen2018_electronlossboost12x', '1M$_\odot$ (Shen+18) 12x e- loss'),
                ]:
            xlist, ylist = amodels[amodelname]
            color = args.color[aindex] if aindex < len(args.color) else None
            print(amodelname, xlist, ylist)
            axis.plot(xlist, ylist, color=color, label='Flörs ' + alabel, marker='+',
                      markersize=10, markeredgewidth=2, lw=0, alpha=0.8)
            aindex += 1

    m18_tdays = np.array([206, 229, 303, 339])
    m18_pew = {}
    # m18_pew[(26, 2, 12570)] = np.array([2383, 1941, 2798, 6770])
    m18_pew[(26, 2, 7155)] = np.array([618, 417, 406, 474])
    m18_pew[(28, 2, 7378)] = np.array([157, 256, 236, 309])
    if args.emfeaturesearch[1][:3] in m18_pew and args.emfeaturesearch[0][:3] in m18_pew:
        axis.set_ylim(ymax=12)
        arr_fratio = m18_pew[args.emfeaturesearch[1][:3]] / m18_pew[args.emfeaturesearch[0][:3]]
        for ax in axes:
            ax.plot(m18_tdays, arr_fratio, color='black', label='Maguire et al. (2018)', lw=2., marker='s')

    for ax in axes:
        ax.set_xlabel(r'Time [days]')
        if not args.nolegend:
            ax.legend(loc='upper right', frameon=False, handlelength=1, ncol=2, numpoints=1, prop={'size': 9})

    defaultoutputfile = 'linefluxes.pdf'
    if not args.outputfile:
        args.outputfile = defaultoutputfile
    elif not Path(args.outputfile).suffixes:
        args.outputfile = args.outputfile / defaultoutputfile

    fig.savefig(args.outputfile, format='pdf')
    # plt.show()
    print(f'Saved {args.outputfile}')
    plt.close()


@at.diskcache(savezipped=True)
def get_packets_with_emission_conditions(modelpath, emtypecolumn, lineindices, tstart, tend, maxpacketfiles=None):
    estimators = at.estimators.read_estimators(modelpath, get_ion_values=False, get_heatingcooling=False)

    modeldata, _, _ = at.inputmodel.get_modeldata(modelpath)
    ts = at.get_timestep_of_timedays(modelpath, tend)
    allnonemptymgilist = [modelgridindex for modelgridindex in modeldata.index
                          if not estimators[(ts, modelgridindex)]['emptycell']]

    # model_tmids = at.get_timestep_times_float(modelpath, loc='mid')
    # arr_velocity_mid = tuple(list([(float(v1) + float(v2)) * 0.5 for v1, v2 in zip(
    #     modeldata['velocity_inner'].values, modeldata['velocity_outer'].values)]))

    # from scipy.interpolate import interp1d
    # interp_log10nne, interp_te = {}, {}
    # for ts in range(len(model_tmids)):
    #     arr_v = np.zeros_like(allnonemptymgilist, dtype='float')
    #     arr_log10nne = np.zeros_like(allnonemptymgilist, dtype='float')
    #     arr_te = np.zeros_like(allnonemptymgilist, dtype='float')
    #     for i, mgi in enumerate(allnonemptymgilist):
    #         arr_v[i] = arr_velocity_mid[mgi]
    #         arr_log10nne[i] = math.log10(float(estimators[(ts, mgi)]['nne']))
    #         arr_te[i] = estimators[(ts, mgi)]['Te']
    #
    #     interp_log10nne[ts] =interp1d(arr_v.copy(), arr_log10nne.copy(),
    #                                                kind='linear', fill_value='extrapolate')
    #     interp_te[ts] = interp1d(arr_v.copy(), arr_te.copy(), kind='linear', fill_value='extrapolate')

    em_mgicolumn = 'em_modelgridindex' if emtypecolumn == 'emissiontype' else 'emtrue_modelgridindex'

    dfpackets_selected, _ = get_packets_with_emtype(
        modelpath, emtypecolumn, lineindices, maxpacketfiles=maxpacketfiles)

    dfpackets_selected = dfpackets_selected.query(
        't_arrive_d >= @tstart and t_arrive_d <= @tend', inplace=False).copy()

    dfpackets_selected = at.packets.add_derived_columns(dfpackets_selected, modelpath, ['em_timestep', em_mgicolumn],
                                                        allnonemptymgilist=allnonemptymgilist)

    if not dfpackets_selected.empty:
        def em_lognne(packet):
            # return interp_log10nne[packet.em_timestep](packet.true_emission_velocity)
            return math.log10(estimators[(int(packet['em_timestep']), int(packet[em_mgicolumn]))]['nne'])

        dfpackets_selected['em_log10nne'] = dfpackets_selected.apply(em_lognne, axis=1)

        def em_Te(packet):
            # return interp_te[packet.em_timestep](packet.true_emission_velocity)
            return estimators[(int(packet['em_timestep']), int(packet[em_mgicolumn]))]['Te']

        dfpackets_selected['em_Te'] = dfpackets_selected.apply(em_Te, axis=1)

    return dfpackets_selected


def plot_nne_te_points(axis, serieslabel, em_log10nne, em_Te, normtotalpackets, color, marker='o'):
    # color_adj = [(c + 0.3) / 1.3 for c in mpl.colors.to_rgb(color)]
    color_adj = [(c + 0.1) / 1.1 for c in mpl.colors.to_rgb(color)]
    hitcount = {}
    for log10nne, Te in zip(em_log10nne, em_Te):
        hitcount[(log10nne, Te)] = hitcount.get((log10nne, Te), 0) + 1

    if hitcount:
        arr_log10nne, arr_te = zip(*hitcount.keys())
    else:
        arr_log10nne, arr_te = np.array([]), np.array([])

    arr_weight = np.array([hitcount[(x, y)] for x, y in zip(arr_log10nne, arr_te)])
    arr_weight = (arr_weight / normtotalpackets) * 500
    arr_size = np.sqrt(arr_weight) * 10

    # arr_weight = arr_weight / float(max(arr_weight))
    # arr_color = np.zeros((len(arr_x), 4))
    # arr_color[:, :3] = np.array([[c for c in mpl.colors.to_rgb(color)] for x in arr_weight])
    # arr_color[:, 3] = (arr_weight + 0.2) / 1.2
    # np.array([[c * z for c in mpl.colors.to_rgb(color)] for z in arr_z])

    # axis.scatter(arr_log10nne, arr_te, s=arr_weight * 20, marker=marker, color=color_adj, lw=0, alpha=1.0,
    #              edgecolors='none')
    alpha = 0.8
    axis.scatter(arr_log10nne, arr_te, s=arr_size, marker=marker, color=color_adj, lw=0, alpha=alpha)

    # make an invisible plot series to appear in the legend with a fixed marker size
    axis.plot([0], [0], marker=marker, markersize=3, color=color_adj, linestyle='None', label=serieslabel, alpha=alpha)

    # axis.plot(em_log10nne, em_Te, label=serieslabel, linestyle='None',
    #           marker='o', markersize=2.5, markeredgewidth=0, alpha=0.05,
    #           fillstyle='full', color=color_b)


def plot_nne_te_bars(axis, serieslabel, em_log10nne, em_Te, color):
    if len(em_log10nne) == 0:
        return
    errorbarkwargs = dict(xerr=np.std(em_log10nne), yerr=np.std(em_Te),
                          color='black', markersize=10., fillstyle='full',
                          capthick=4, capsize=15, linewidth=4.,
                          alpha=1.0)
    # black larger one for an outline
    axis.errorbar(np.mean(em_log10nne), np.mean(em_Te), **errorbarkwargs)
    errorbarkwargs['markersize'] -= 2
    errorbarkwargs['capthick'] -= 2
    errorbarkwargs['capsize'] -= 1
    errorbarkwargs['linewidth'] -= 2
    errorbarkwargs['color'] = color
    # errorbarkwargs['zorder'] += 0.5
    axis.errorbar(np.mean(em_log10nne), np.mean(em_Te), **errorbarkwargs)


def make_emitting_regions_plot(args):
    import artistools.estimators

    # font = {'size': 16}
    # matplotlib.rc('font', **font)
    # 'floers_te_nne.json',
    refdatafilenames = ['floers_te_nne.json', ]  # , 'floers_te_nne_CMFGEN.json', 'floers_te_nne_Smyth.json']
    refdatalabels = ['Flörs+2020', ]  # , 'Floers CMFGEN', 'Floers Smyth']
    refdatacolors = ['0.0', 'C1', 'C2', 'C4']
    refdatakeys = [None for _ in refdatafilenames]
    refdatatimes = [None for _ in refdatafilenames]
    refdatapoints = [None for _ in refdatafilenames]
    for refdataindex, refdatafilename in enumerate(refdatafilenames):
        with open(refdatafilename, encoding='utf-8') as data_file:
            floers_te_nne = json.loads(data_file.read())

        # give an ordering and index to dict items
        refdatakeys[refdataindex] = [t for t in sorted(floers_te_nne.keys(), key=lambda x: float(x))]  # strings, not floats
        refdatatimes[refdataindex] = np.array([float(t) for t in refdatakeys[refdataindex]])
        refdatapoints[refdataindex] = [floers_te_nne[t] for t in refdatakeys[refdataindex]]
        print(f'{refdatafilename} data available for times: {list(refdatatimes[refdataindex])}')

    times_days = (np.array(args.timebins_tstart) + np.array(args.timebins_tend)) / 2.

    print(f'Chosen times: {times_days}')

    # axis.set_xlim(left=supxmin, right=supxmax)
    # pd.set_option('display.max_rows', 50)
    pd.set_option('display.width', 250)
    pd.options.display.max_rows = 500

    emdata_all = {}
    log10nnedata_all = {}
    Tedata_all = {}

    # data is collected, now make plots
    defaultoutputfile = 'emittingregions.pdf'
    if not args.outputfile:
        args.outputfile = defaultoutputfile
    elif not Path(args.outputfile).suffixes:
        args.outputfile = args.outputfile / defaultoutputfile

    args.modelpath.append(None)
    args.label.append(f'All models: {args.label}')
    args.modeltag.append('all')
    for modelindex, (modelpath, modellabel, modeltag) in enumerate(
            zip(args.modelpath, args.label, args.modeltag)):

        print(f"ARTIS model: '{modellabel}'")

        if modelpath is not None:
            print(f"Getting packets/nne/Te data for ARTIS model: '{modellabel}'")

            emdata_all[modelindex] = {}

            emfeatures = get_labelandlineindices(modelpath, tuple(args.emfeaturesearch))

            linelistindices_allfeatures = tuple([l for feature in emfeatures for l in feature.linelistindices])

            for tmid, tstart, tend in zip(times_days, args.timebins_tstart, args.timebins_tend):
                dfpackets = get_packets_with_emission_conditions(
                    modelpath, args.emtypecolumn, linelistindices_allfeatures, tstart, tend, maxpacketfiles=args.maxpacketfiles)

                for feature in emfeatures:
                    dfpackets_selected = dfpackets.query(f'{args.emtypecolumn} in @feature.linelistindices', inplace=False)
                    if dfpackets_selected.empty:
                        emdata_all[modelindex][(tmid, feature.colname)] = {
                            'em_log10nne': [],
                            'em_Te': []}
                    else:
                        emdata_all[modelindex][(tmid, feature.colname)] = {
                            'em_log10nne': dfpackets_selected.em_log10nne.values,
                            'em_Te': dfpackets_selected.em_Te.values}

            estimators = at.estimators.read_estimators(modelpath, get_ion_values=False, get_heatingcooling=False)
            modeldata, _, _ = at.inputmodel.get_modeldata(modelpath)
            Tedata_all[modelindex] = {}
            log10nnedata_all[modelindex] = {}
            for tmid, tstart, tend in zip(times_days, args.timebins_tstart, args.timebins_tend):
                Tedata_all[modelindex][tmid] = []
                log10nnedata_all[modelindex][tmid] = []
                tstartlist = at.get_timestep_times_float(modelpath, loc='start')
                tendlist = at.get_timestep_times_float(modelpath, loc='end')
                tslist = [ts for ts in range(len(tstartlist)) if tendlist[ts] >= tstart and tstartlist[ts] <= tend]
                for timestep in tslist:
                    for modelgridindex in modeldata.index:
                        try:
                            Tedata_all[modelindex][tmid].append(estimators[(timestep, modelgridindex)]['Te'])
                            log10nnedata_all[modelindex][tmid].append(math.log10(estimators[(timestep, modelgridindex)]['nne']))
                        except KeyError:
                            pass

        if modeltag != 'all':
            continue

        for tmid in times_days:
            print(f'  Plot at {tmid} days')

            nrows = 1
            fig, axis = plt.subplots(
                nrows=nrows, ncols=1, sharey=False, sharex=False,
                figsize=(args.figscale * at.figwidth, args.figscale * at.figwidth * (0.25 + nrows * 0.7)),
                tight_layout={"pad": 0.2, "w_pad": 0.0, "h_pad": 0.2})

            for refdataindex, f in enumerate(refdatafilenames):
                timeindex = np.abs(refdatatimes[refdataindex] - tmid).argmin()
                axis.plot(refdatapoints[refdataindex][timeindex]['ne'], refdatapoints[refdataindex][timeindex]['temp'],
                          color=refdatacolors[refdataindex], lw=2, label=f'{refdatalabels[refdataindex]} +{refdatakeys[refdataindex][timeindex]}d')

                timeindexb = np.abs(refdatatimes[refdataindex] - tmid - 50).argmin()
                if timeindexb < len(refdatakeys[refdataindex]):
                    axis.plot(refdatapoints[refdataindex][timeindexb]['ne'], refdatapoints[refdataindex][timeindexb]['temp'],
                              color='0.4', lw=2, label=f'{refdatalabels[refdataindex]} +{refdatakeys[refdataindex][timeindexb]}d')

            if modeltag == 'all':
                for bars in [False, ]:  # [False, True]
                    for truemodelindex in range(modelindex):
                        emfeatures = get_labelandlineindices(args.modelpath[truemodelindex], args.emfeaturesearch)

                        em_log10nne = np.concatenate(
                            [emdata_all[truemodelindex][(tmid, feature.colname)]['em_log10nne']
                             for feature in emfeatures])

                        em_Te = np.concatenate(
                            [emdata_all[truemodelindex][(tmid, feature.colname)]['em_Te']
                             for feature in emfeatures])

                        normtotalpackets = len(em_log10nne) * 8.  # circles have more area than triangles, so decrease
                        modelcolor = args.color[truemodelindex]
                        label = args.label[truemodelindex].format(timeavg=tmid, modeltag=modeltag)
                        if not bars:
                            plot_nne_te_points(
                                axis, label, em_log10nne, em_Te, normtotalpackets, modelcolor)
                        else:
                            plot_nne_te_bars(axis, args.label[truemodelindex], em_log10nne, em_Te, modelcolor)
            else:
                modellabel = args.label[modelindex]
                emfeatures = get_labelandlineindices(modelpath, tuple(args.emfeaturesearch))

                featurecolours = ['blue', 'red']
                markers = [10, 11]
                # featurecolours = ['C0', 'C3']
                # featurebarcolours = ['blue', 'red']

                normtotalpackets = np.sum([len(emdata_all[modelindex][(tmid, feature.colname)]['em_log10nne'])
                                           for feature in emfeatures])

                axis.scatter(log10nnedata_all[modelindex][tmid], Tedata_all[modelindex][tmid], s=1.0, marker='o',
                             color='0.4', lw=0, edgecolors='none', label='All cells')

                for bars in [False, ]:  # [False, True]
                    for featureindex, feature in enumerate(emfeatures):
                        emdata = emdata_all[modelindex][(tmid, feature.colname)]

                        if not bars:
                            print(f'   {len(emdata["em_log10nne"])} points plotted for {feature.featurelabel}')

                        serieslabel = (
                            modellabel + ' ' + feature.featurelabel).format(
                            timeavg=tmid, modeltag=modeltag).replace('Å', r' $\mathrm{\AA}$')

                        if not bars:
                            plot_nne_te_points(
                                axis, serieslabel, emdata['em_log10nne'], emdata['em_Te'],
                                normtotalpackets, featurecolours[featureindex], marker=markers[featureindex])
                        else:
                            plot_nne_te_bars(
                                axis, serieslabel, emdata['em_log10nne'], emdata['em_Te'], featurecolours[featureindex])

            if tmid == times_days[-1] and not args.nolegend:
                axis.legend(loc='best', frameon=False, handlelength=1, ncol=1, borderpad=0,
                            numpoints=1, fontsize=11, markerscale=2.5)

            axis.set_ylim(ymin=3000)
            axis.set_ylim(ymax=10000)
            axis.set_xlim(xmin=4.5, xmax=7.15)

            axis.set_xlabel(r'log$_{10}$(n$_{\mathrm{e}}$ [cm$^{-3}$])')
            axis.set_ylabel(r'Electron Temperature [K]')

            # axis.annotate(f'{tmid:.0f}d', xy=(0.98, 0.5), xycoords='axes fraction',
            #               horizontalalignment='right', verticalalignment='center', fontsize=16)

            outputfile = str(args.outputfile).format(timeavg=tmid, modeltag=modeltag)
            fig.savefig(outputfile, format='pdf')
            print(f'    Saved {outputfile}')
            plt.close()


def addargs(parser):
    parser.add_argument('-modelpath', default=[], nargs='*', action=at.AppendPath,
                        help='Paths to ARTIS folders with spec.out or packets files')

    parser.add_argument('-label', default=[], nargs='*',
                        help='List of series label overrides')

    parser.add_argument('--nolegend', action='store_true',
                        help='Suppress the legend from the plot')

    parser.add_argument('-modeltag', default=[], nargs='*',
                        help='List of model tags for file names')

    parser.add_argument('-color', default=[f'C{i}' for i in range(10)], nargs='*',
                        help='List of line colors')

    parser.add_argument('-linestyle', default=[], nargs='*',
                        help='List of line styles')

    parser.add_argument('-linewidth', default=[], nargs='*',
                        help='List of line widths')

    parser.add_argument('-dashes', default=[], nargs='*',
                        help='Dashes property of lines')

    parser.add_argument('-maxpacketfiles', type=int, default=None,
                        help='Limit the number of packet files read')

    parser.add_argument('-emfeaturesearch', default=[], nargs='*',
                        help='List of tuples (TODO explain)')

    # parser.add_argument('-emtypecolumn', default='trueemissiontype', choices=['emissiontype', 'trueemissiontype'],
    #                     help='Packet property for emission type - first thermal emission (trueemissiontype) '
    #                     'or last emission type (emissiontype)')

    parser.add_argument('--frompops', action='store_true',
                        help='Sum up internal emissivity instead of outgoing packets')

    parser.add_argument('--use_lastemissiontype', action='store_true',
                        help='Tag packets by their last scattering rather than thermal emission type')

    # parser.add_argument('-timemin', type=float,
    #                     help='Lower time in days to integrate spectrum')
    #
    # parser.add_argument('-timemax', type=float,
    #                     help='Upper time in days to integrate spectrum')
    #
    parser.add_argument('-xmin', type=int, default=50,
                        help='Plot range: minimum wavelength in Angstroms')

    parser.add_argument('-xmax', type=int, default=450,
                        help='Plot range: maximum wavelength in Angstroms')

    parser.add_argument('-ymin', type=float, default=None,
                        help='Plot range: y-axis')

    parser.add_argument('-ymax', type=float, default=None,
                        help='Plot range: y-axis')

    parser.add_argument('-timebins_tstart', default=[], nargs='*', action='append',
                        help='Time bin start values in days')

    parser.add_argument('-timebins_tend', default=[], nargs='*', action='append',
                        help='Time bin end values in days')

    parser.add_argument('-figscale', type=float, default=1.8,
                        help='Scale factor for plot area. 1.0 is for single-column')

    parser.add_argument('--write_data', action='store_true',
                        help='Save data used to generate the plot in a CSV file')

    parser.add_argument('--plotemittingregions', action='store_true',
                        help='Plot conditions where flux line is emitted')

    parser.add_argument('-outputfile', '-o', action='store', dest='outputfile', type=Path,
                        help='path/filename for PDF file')


def main(args=None, argsraw=None, **kwargs):
    """Plot spectra from ARTIS and reference data."""
    if args is None:
        parser = argparse.ArgumentParser(
            formatter_class=at.CustomArgHelpFormatter,
            description='Plot ARTIS model spectra by finding spec.out files '
                        'in the current directory or subdirectories.')
        addargs(parser)
        parser.set_defaults(**kwargs)
        args = parser.parse_args(argsraw)

    if not args.modelpath:
        args.modelpath = [Path('.')]
    elif isinstance(args.modelpath, (str, Path)):
        args.modelpath = [args.modelpath]

    args.modelpath = at.flatten_list(args.modelpath)

    args.label, args.modeltag, args.color = at.trim_or_pad(len(args.modelpath), args.label, args.modeltag, args.color)

    args.emtypecolumn = 'emissiontype' if args.use_lastemissiontype else 'trueemissiontype'

    for i in range(len(args.label)):
        if args.label[i] is None:
            args.label[i] = at.get_model_name(args.modelpath[i])

    if args.plotemittingregions:
        make_emitting_regions_plot(args)
    else:
        make_flux_ratio_plot(args)


if __name__ == "__main__":
    multiprocessing.freeze_support()
    main()



import psutil
from pathlib import Path

# num_processes = 1
num_processes = max(1, int(psutil.cpu_count(logical=False)) - 2)
# print(f'Using {num_processes} processes')

enable_diskcache = True

figwidth = 5

config = {}
config['codecomparisondata1path'] = Path(
    '/Users/luke/Library/Mobile Documents/com~apple~CloudDocs/GitHub/sn-rad-trans/data1')

config['codecomparisonmodelartismodelpath'] = Path('/Volumes/GoogleDrive/My Drive/artis_runs/weizmann/')

config['path_artistools_repository'] = Path(__file__).absolute().parent.parent
config['path_artistools_sourcedir'] = Path(__file__).absolute().parent
config['path_datadir'] = Path(__file__).absolute().parent / 'data'
config['path_testartismodel'] = Path(config['path_artistools_repository'], 'tests', 'data', 'testmodel')
config['path_testoutput'] = Path(config['path_artistools_repository'], 'tests', 'output')

config['gsimerger_trajroot'] = Path(
    '/Users/luke/Library/Mobile Documents/com~apple~CloudDocs/Archive/Astronomy/Mergers/SFHo/')



import tarfile
from pathlib import Path
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from scipy.integrate import quad
import os

import artistools as at
import artistools.inputmodel


DAY = 86400  # day in seconds
MSUN = 1.989e33  # solar mass in grams


def write_energydistribution_file(energydistdata, outputfilepath='.'):
    print('Writing energydistribution.txt')
    with open(Path(outputfilepath) / 'energydistribution.txt', 'w') as fmodel:
        fmodel.write(f'{len(energydistdata["cell_energy"])}\n')  # write number of points
        energydistdata.to_csv(fmodel, header=False, sep='\t', index=False, float_format='%g')


def write_energyrate_file(energy_rate_data, outputfilepath='.'):
    print('Writing energyrate.txt')
    with open(Path(outputfilepath) / 'energyrate.txt', 'w') as fmodel:
        fmodel.write(f'{len(energy_rate_data["times"])}\n')  # write number of points
        energy_rate_data.to_csv(fmodel, sep='\t', index=False, header=False, float_format='%.10f')


def rprocess_const_and_powerlaw():
    """Following eqn 4 Korobkin 2012"""

    def integrand(t_days, t0, epsilon0, sigma, alpha, thermalisation_factor):
        return (epsilon0 * ((1/2) - (1/np.pi * np.arctan((t_days-t0)/sigma)))**alpha) * (thermalisation_factor / 0.5)

    tmin = 0.01*DAY
    tmax = 50*DAY
    t0 = 1.3  # seconds
    epsilon0 = 2e18
    sigma = 0.11
    alpha = 1.3
    thermalisation_factor = 0.5

    E_tot = quad(integrand, tmin, tmax, args=(t0, epsilon0, sigma, alpha, thermalisation_factor))  # ergs/s/g
    print("Etot per gram", E_tot[0])
    E_tot = E_tot[0]

    times = np.logspace(np.log10(tmin), np.log10(tmax), num=200)
    energy_per_gram_cumulative = [0]
    for time in times[1:]:
        cumulative_integral = (quad(integrand, tmin, time, args=(t0, epsilon0, sigma, alpha, thermalisation_factor)))  # ergs/s/g
        energy_per_gram_cumulative.append(cumulative_integral[0])

    energy_per_gram_cumulative = np.array(energy_per_gram_cumulative)

    rate = energy_per_gram_cumulative / E_tot

    nuclear_heating_power = []
    for time in times:
        nuclear_heating_power.append(integrand(time, t0, epsilon0, sigma, alpha, thermalisation_factor))

    # times_and_rate = {'times': times/DAY, 'rate': rate, 'nuclear_heating_power': nuclear_heating_power}
    times_and_rate = {'times': times/DAY, 'rate': rate}
    times_and_rate = pd.DataFrame(data=times_and_rate)

    return times_and_rate, E_tot


def energy_from_rprocess_calculation(energy_thermo_data, get_rate=True):

    index_time_greaterthan = energy_thermo_data[energy_thermo_data['time/s'] > 1e7].index  # 1e7 seconds = 116 days
    energy_thermo_data.drop(index_time_greaterthan, inplace=True)
    # print("Dropping times later than 116 days")

    skipfirstnrows = 0  # not sure first values look sensible -- check this
    times = energy_thermo_data['time/s'][skipfirstnrows:]
    qdot = energy_thermo_data['Qdot'][skipfirstnrows:]

    E_tot = np.trapz(y=qdot, x=times)  # erg / g

    if get_rate:
        print(f"E_tot {E_tot} erg/g")

        import scipy.integrate
        cumulative_integrated_energy = scipy.integrate.cumulative_trapezoid(y=qdot, x=times)
        cumulative_integrated_energy = np.insert(cumulative_integrated_energy, 0, 0)

        rate = cumulative_integrated_energy / E_tot

        times_and_rate = {'times': times/DAY, 'rate': rate}
        times_and_rate = pd.DataFrame(data=times_and_rate)

        return times_and_rate, E_tot

    else:
        return E_tot


def get_rprocess_calculation_files(path_to_rprocess_calculation, interpolate_trajectories=False, thermalisation=False):
    tarfiles = [file for file in os.listdir(path_to_rprocess_calculation) if file.endswith(".tar.xz")]

    trajectory_ids = []
    trajectory_E_tot = []

    if interpolate_trajectories:
        interpolated_trajectories = {'time/s': np.logspace(-1, 7, 300)}

    energy_thermo_filepath = "./Run_rprocess/energy_thermo.dat"
    for file in tarfiles:
        trajectory_id = file.split('.')[0]
        tar = tarfile.open(path_to_rprocess_calculation / file, mode='r:*')

        energythermo_file = tar.extractfile(member=energy_thermo_filepath)
        energy_thermo_data = pd.read_csv(energythermo_file, delim_whitespace=True)
        # print(energy_thermo_data['Qdot'])
        # print(energy_thermo_data['time/s'])

        if interpolate_trajectories:
            qdotinterp = np.interp(interpolated_trajectories['time/s'], energy_thermo_data['time/s'],
                                   energy_thermo_data['Qdot'])
            interpolated_trajectories[trajectory_id] = qdotinterp

        E_tot = energy_from_rprocess_calculation(energy_thermo_data, get_rate=False, thermalisation=thermalisation)

        trajectory_ids.append(float(trajectory_id))
        trajectory_E_tot.append(E_tot)

    if interpolate_trajectories:
        interpolated_trajectories = pd.DataFrame.from_dict(interpolated_trajectories)
        interpolated_trajectories['mean'] = interpolated_trajectories.iloc[:, 1:].mean(axis=1)

        index_time_lessthan = interpolated_trajectories[interpolated_trajectories['time/s'] < 1.1e-1].index
        interpolated_trajectories.drop(index_time_lessthan, inplace=True)

        interpolated_trajectories.to_csv(path_to_rprocess_calculation / 'interpolatedQdot.dat', sep=' ', index=False)
    print(f"sum etot {sum(trajectory_E_tot)}")
    trajectory_energy = {'id': trajectory_ids, 'E_tot': trajectory_E_tot}
    trajectory_energy = pd.DataFrame.from_dict(trajectory_energy)
    trajectory_energy = trajectory_energy.sort_values(by='id')
    print(trajectory_energy)
    trajectory_energy.to_csv(path_to_rprocess_calculation / 'trajectoryQ.dat', sep=' ', index=False)


def make_energydistribution_weightedbyrho(rho, E_tot_per_gram, Mtot_grams):

    Etot = E_tot_per_gram * Mtot_grams
    print("Etot", Etot)
    numberofcells = len(rho)

    cellenergy = np.array([Etot] * numberofcells)
    cellenergy = cellenergy * (rho / sum(rho))

    energydistdata = {
        'cellid': np.arange(1, len(rho)+1),
        'cell_energy': cellenergy
    }

    print("sum energy cells", sum(energydistdata['cell_energy']))
    energydistdata = pd.DataFrame(data=energydistdata)

    return energydistdata


def make_energy_files(rho, Mtot_grams, outputpath=None):
    times_and_rate, E_tot_per_gram = rprocess_const_and_powerlaw()
    energydistributiondata = make_energydistribution_weightedbyrho(rho, E_tot_per_gram, Mtot_grams)

    write_energydistribution_file(energydistributiondata, outputfilepath=outputpath)
    write_energyrate_file(times_and_rate, outputfilepath=outputpath)


def plot_energy_rate(modelpath):
    times_and_rate, E_tot = at.inputmodel.energyinputfiles.rprocess_const_and_powerlaw()
    model, _, _ = at.inputmodel.get_modeldata(modelpath)
    Mtot_grams = model['cellmass_grams'].sum()
    plt.plot(times_and_rate['times'], np.array(times_and_rate['nuclear_heating_power'])*Mtot_grams,
             color='k', zorder=10)



#!/usr/bin/env python3

import argparse
import math
# import os.path

import numpy as np
# import pandas as pd

from artistools import CustomArgHelpFormatter, get_atomic_number
import artistools.inputmodel


def addargs(parser):

    parser.add_argument('-inputfile', '-i', default='model.txt',
                        help='Path of input file or folder containing model.txt')

    parser.add_argument('-cell', '-mgi', default=None,
                        help='Focus on particular cell number (0-indexed)')

    parser.add_argument('--noabund', action='store_true',
                        help='Give total masses only, no nuclear or elemental abundances')

    parser.add_argument('--getabundances', action='store_true',
                        help='Get elemental abundance masses')


def main(args=None, argsraw=None, **kwargs):
    if args is None:
        parser = argparse.ArgumentParser(
            formatter_class=CustomArgHelpFormatter,
            description='Scale the velocity of an ARTIS model, keeping mass constant and saving back to ARTIS format.')

        addargs(parser)
        parser.set_defaults(**kwargs)
        args = parser.parse_args(argsraw)

    print(f'Reading {args.inputfile}')
    dfmodel, t_model_init_days, vmax = artistools.inputmodel.get_modeldata(
        args.inputfile, get_abundances=args.getabundances)

    t_model_init_seconds = t_model_init_days * 24 * 60 * 60
    print(f'Model is defined at {t_model_init_days} days ({t_model_init_seconds:.4f} seconds)')

    if 'pos_x_min' in dfmodel.columns:
        nonemptycells = sum(dfmodel['rho'] > 0.)
        print(f'Model contains {len(dfmodel)} Cartesian grid cells ({nonemptycells} nonempty) with '
              f'vmax = {vmax} km/s ({vmax / 299792.458:.2f} * c)')
        corner_vmax = math.sqrt(3 * vmax ** 2)
        print(f'  corner vmax: {corner_vmax:.2e} cm/s ({corner_vmax / 299792.458:.2f} * c)')
    else:
        vmax = dfmodel['velocity_outer'].max()
        print(f'Model contains {len(dfmodel)} 1D spherical shells with vmax = {vmax} km/s '
              f'({vmax / 299792.458:.2f} * c)')

    if args.cell is not None:
        mgi = int(args.cell)
        if mgi >= 0:
            print(f'Selected single cell mgi {mgi}:')
            dfmodel.query('inputcellid == (@mgi + 1)', inplace=True)
            print(dfmodel.iloc[0])

    mass_msun_rho = dfmodel['cellmass_grams'].sum() / 1.989e33

    mass_msun_isotopes = 0.
    mass_msun_elem = 0.
    speciesmasses = {}
    for column in dfmodel.columns:
        if column.startswith('X_'):
            species = column.replace('X_', '')
            speciesabund_g = np.dot(dfmodel[column], dfmodel['cellmass_grams'])

            species_mass_msun = speciesabund_g / 1.989e33
            atomic_number = get_atomic_number(species)
            if species[-1].isdigit():
                strtotiso = species.rstrip('0123456789') + '-isosum'
                speciesmasses[strtotiso] = speciesmasses.get(strtotiso, 0.) + speciesabund_g
                mass_msun_isotopes += species_mass_msun
            elif species.lower() != 'fegroup':
                mass_msun_elem += species_mass_msun

            if speciesabund_g > 0.:
                speciesmasses[species] = speciesabund_g

    print(f'M_{"tot_rho":9s} {mass_msun_rho:8.5f} MSun (density * volume)')
    if mass_msun_elem > 0.:
        print(f'M_{"tot_elem":9s} {mass_msun_elem:8.5f} MSun ({mass_msun_elem / mass_msun_rho * 100:6.2f}% of M_tot_rho)')

    print(f'M_{"tot_iso":9s} {mass_msun_isotopes:8.5f} MSun ({mass_msun_isotopes / mass_msun_rho * 100:6.2f}% of M_tot_rho, but can be < 100% if stable isotopes not tracked)')

    if not args.noabund:
        for species, mass_g in sorted(speciesmasses.items(), key=lambda x: (get_atomic_number(x[0]), x[0])):
            species_mass_msun = mass_g / 1.989e33
            massfrac = species_mass_msun / mass_msun_rho
            strcomment = ''
            if species.endswith('-isosum') and args.getabundances:
                elsymb = species.replace('-isosum', '')
                elem_mass = speciesmasses.get(elsymb, 0.)
                if elem_mass > 0.:
                    strcomment += f' ({mass_g / elem_mass * 100:6.2f}% of {elsymb} element mass)'
                if mass_g > elem_mass:
                    strcomment += ' ERROR! isotope sum is greater than element abundance'

            print(f'{species:9s} {species_mass_msun:.3e} Msun    massfrac {massfrac:.3e}{strcomment}')


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import argparse
# import os
from pathlib import Path
# import pandas as pd
import artistools as at
import matplotlib.pyplot as plt
from astropy import units as u
# import matplotlib
import numpy as np
# from mpl_toolkits.mplot3d import Axes3D
import gc


def make_cone(args):
    print("Making cone")

    angle_of_cone = 30  # in deg

    theta = np.radians([angle_of_cone / 2])  # angle between line of sight and edge is half angle of cone

    # merge_dfs, args.t_model, args.vmax = at.inputmodel.get_modeldata(args.modelpath[0], dimensions=3, get_abundances=True)
    merge_dfs = at.inputmodel.get_3d_model_data_merged_model_and_abundances_minimal(args)

    if args.positive_axis:
        cone = (merge_dfs.loc[merge_dfs[f'cellpos_in[{args.sliceaxis}]'] >= 1 / (np.tan(theta))
                              * np.sqrt((merge_dfs[f'cellpos_in[{args.other_axis2}]']) ** 2
                                        + (merge_dfs[f'cellpos_in[{args.other_axis1}]']) ** 2)])  # positive axis
    else:
        cone = (merge_dfs.loc[merge_dfs[f'cellpos_in[{args.sliceaxis}]'] <= -1/(np.tan(theta))
                              * np.sqrt((merge_dfs[f'cellpos_in[{args.other_axis2}]'])**2
                                        + (merge_dfs[f'cellpos_in[{args.other_axis1}]'])**2)])  # negative axis
    # print(cone.loc[:, :[f'pos_{slice_on_axis}']])

    del merge_dfs  # merge_dfs not needed anymore so free memory
    gc.collect()

    return cone


def get_profile_along_axis(args=None, modeldata=None, derived_cols=False):
    print("Getting profile along axis")

    # merge_dfs, args.t_model, args.vmax = at.inputmodel.get_modeldata(args.modelpath, dimensions=3, get_abundances=True)
    if modeldata is None:
        modeldata, _, _ = at.inputmodel.get_modeldata(args.modelpath, dimensions=3, get_abundances=True,
                                                      derived_cols=derived_cols)

    position_closest_to_axis = modeldata.iloc[
        (modeldata[f'pos_{args.other_axis2}']).abs().argsort()][:1][f'pos_{args.other_axis2}'].item()

    if args.positive_axis:
        profile1D = modeldata.loc[(modeldata[f'pos_{args.other_axis1}'] == position_closest_to_axis)
                                  & (modeldata[f'pos_{args.other_axis2}'] == position_closest_to_axis)
                                  & (modeldata[f'pos_{args.sliceaxis}'] > 0)]
    else:
        profile1D = modeldata.loc[(modeldata[f'pos_{args.other_axis1}'] == position_closest_to_axis)
                                  & (modeldata[f'pos_{args.other_axis2}'] == position_closest_to_axis)
                                  & (modeldata[f'pos_{args.sliceaxis}'] < 0)]

    return profile1D.reset_index(drop=True)


def make_1D_profile(args):
    if args.makefromcone:
        cone = make_cone(args)

        slice1D = cone.groupby([f'cellpos_in[{args.sliceaxis}]'], as_index=False).mean()
        # where more than 1 X value, average rows eg. (1,0,0) (1,1,0) (1,1,1)

    else:  # make from along chosen axis
        slice1D = get_profile_along_axis(args)

    slice1D[f'cellpos_in[{args.sliceaxis}]'] = slice1D[f'cellpos_in[{args.sliceaxis}]'].apply(
        lambda x: x / args.t_model * (u.cm / u.day).to('km/s'))  # Convert positions to velocities
    slice1D = slice1D.rename(columns={f'cellpos_in[{args.sliceaxis}]': 'vout_kmps'})
    # Convert position to velocity

    slice1D = slice1D.drop(['inputcellid', f'cellpos_in[{args.other_axis1}]', f'cellpos_in[{args.other_axis2}]'],
                           axis=1)  # Remove columns we don't need

    slice1D['rho_model'] = slice1D['rho_model'].apply(lambda x: np.log10(x) if x != 0 else -100)
    # slice1D = slice1D[slice1D['rho_model'] != -100]  # Remove empty cells
    # TODO: fix this, -100 probably breaks things if it's not one of the outer cells that gets chopped
    slice1D = slice1D.rename(columns={'rho_model': 'log_rho'})

    slice1D.index += 1

    if not args.positive_axis:
        # Invert rows and *velocity by -1 to make velocities positive for slice on negative axis
        slice1D.iloc[:] = slice1D.iloc[::-1].values
        slice1D['vout_kmps'] = slice1D['vout_kmps'].apply(lambda x: x*-1)

    # with pd.option_context('display.max_rows', None, 'display.max_columns', None):
    #     print(slice1D)

    # print(slice1D.keys())
    return slice1D


def make_1D_model_files(args):
    slice1D = make_1D_profile(args)
    query_abundances_positions = slice1D.columns.str.startswith('X_')
    model_df = slice1D.loc[:, ~query_abundances_positions]
    abundances_df = slice1D.loc[:, query_abundances_positions]

    # with pd.option_context('display.max_rows', None, 'display.max_columns', None):  # Print all rows in df
    #     print(model_df)

    # print(modelpath)
    model_df = model_df.round(decimals=5)  # model files seem to be to 5 sf
    model_df.to_csv(args.modelpath[0]/"model_1D.txt", sep=' ', header=False)  # write model.txt

    abundances_df.to_csv(args.modelpath[0]/"abundances_1D.txt", sep=' ', header=False)  # write abundances.txt

    with open(args.modelpath[0]/"model_1D.txt", 'r+') as f:  # add number of cells and tmodel to start of file
        content = f.read()
        f.seek(0, 0)
        f.write(f"{model_df.shape[0]}\n{args.t_model}".rstrip('\r\n') + '\n' + content)

    print("Saved abundances_1D.txt and model_1D.txt")

# with open(args.modelpath[0]/"model

# print(cone)

# cone = (merge_dfs.loc[merge_dfs[f'pos_{args.other_axis2}'] <= - (1/(np.tan(theta))
# * np.sqrt((merge_dfs[f'pos_{slice_on_axis}'])**2 + (merge_dfs[f'pos_{args.other_axis1}'])**2))])
# cone = merge_dfs
# cone = cone.loc[cone['rho_model'] > 0.0]


def make_plot(args):
    cone = make_cone(args)

    cone = cone.loc[cone['rho_model'] > 0.0002]  # cut low densities (empty cells?) from plot
    fig = plt.figure()
    ax = fig.gca(projection='3d')

    # print(cone['rho_model'])

    # set up for big model. For scaled down artis input model switch x and z
    x = cone['pos_z_min'].apply(lambda x: x / args.t_model * (u.cm / u.day).to('km/s'))/1e3
    y = cone['pos_y_min'].apply(lambda x: x / args.t_model * (u.cm / u.day).to('km/s'))/1e3
    z = cone['pos_x_min'].apply(lambda x: x / args.t_model * (u.cm / u.day).to('km/s'))/1e3

    surf = ax.scatter3D(x, y, z, c=-cone['fni'], cmap=plt.get_cmap('viridis'))

    # fig.colorbar(surf, shrink=0.5, aspect=5)

    ax.set_xlabel(r'x [10$^3$ km/s]')
    ax.set_ylabel(r'y [10$^3$ km/s]')
    ax.set_zlabel(r'z [10$^3$ km/s]')

    # plt.scatter(cone[f'pos_x_min']/1e11, cone[f'pos_y_min']/1e11)
    plt.show()


def addargs(parser):
    parser.add_argument('-modelpath', default=[], nargs='*', action=at.AppendPath,
                        help='Path to ARTIS model folders with model.txt and abundances.txt')

    parser.add_argument('-sliceaxis', type=str, default='x',
                        help='Choose x, y or z. The 1D model will be made based on this axis.'
                             'If cone then cone made around sliceaxis. Otherwise model along axis.')

    parser.add_argument('--positive_axis', action='store', default=True,
                        help='Make 1D model from positive axis. Default is True')

    parser.add_argument('--makefromcone', action='store', default=True,
                        help='Make 1D model from cone around axis. Default is True.'
                             'If False uses points along axis.')


def main(args=None, argsraw=None, **kwargs):
    if args is None:
        parser = argparse.ArgumentParser(
            formatter_class=at.CustomArgHelpFormatter,
            description='Make 1D model from cone in 3D model. Call with "makeartismodel1dslicefromcone"')
        addargs(parser)
        parser.set_defaults(**kwargs)
        args = parser.parse_args(argsraw)

    if not args.modelpath:
        args.modelpath = [Path('.')]

    args.other_axis1 = None
    args.other_axis2 = None
    axes = ['x', 'y', 'z']
    for ax in axes:
        if args.other_axis1 is None and ax != args.sliceaxis:
            args.other_axis1 = ax
        elif args.other_axis2 is None and ax != args.sliceaxis:
            args.other_axis2 = ax
    # remember: models before scaling down to artis input have x and z axis swapped compared to artis input files

    make_1D_model_files(args)

    # make_plot(args) # Uncomment to make 3D plot todo: add command line option


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

# coding: utf-8
"""Plotting and analysis tools for the ARTIS 3D supernova radiative transfer code."""

import datetime
import sys

from pathlib import Path
from setuptools import find_packages, setup
from setuptools.command.test import test as TestCommand

# sys.path.append('artistools/')
# from commands import console_scripts
from artistools.commands import console_scripts, completioncommands


# Add the following lines to your .zshrc file to get command completion:
# autoload -U bashcompinit
# bashcompinit
# source artistoolscompletions.sh
with open('artistoolscompletions.sh', 'w') as f:
    f.write('\n'.join(completioncommands))


class PyTest(TestCommand):
    """Setup the py.test test runner."""

    def finalize_options(self):
        """Set options for the command line."""
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        """Execute the test runner command."""
        # Import here, because outside the required eggs aren't loaded yet
        import pytest
        sys.exit(pytest.main(self.test_args))


def get_version():
    utcnow = datetime.datetime.utcnow()
    strversion = f'{utcnow.year}.{utcnow.month}.{utcnow.day}.'
    strversion += f'{utcnow.hour:02d}{utcnow.minute:02d}{utcnow.second:02d}.'
    strversion += 'dev'
    return strversion


setup(
    name="artistools",
    version='2022.04.04',
    author="ARTIS Collaboration",
    author_email="luke.shingles@gmail.com",
    packages=find_packages(),
    url="https://www.github.com/artis-mcrt/artistools/",
    license="MIT",
    description="Plotting and analysis tools for the ARTIS 3D supernova radiative transfer code.",
    long_description=(Path(__file__).absolute().parent / "README.md").open('rt').read(),
    long_description_content_type='text/markdown',
    install_requires=(Path(__file__).absolute().parent / "requirements.txt").open('rt').read().splitlines(),
    entry_points={
        'console_scripts': console_scripts,
    },
    scripts=['artistoolscompletions.sh'],
    python_requires='>==3.6',
    # test_suite='tests',
    setup_requires=['pytest', 'pytest-runner', 'pytest-cov'],
    tests_require=['pytest', 'pytest-runner', 'pytest-cov'],
    include_package_data=True)



import numpy as np
# import pandas as pd

import artistools as at

CLIGHT = 2.99792458e10


def change_cell_positions_to_new_time(dfgriddata, t_model_1d):
    dfgriddata['pos_x_min'] = dfgriddata['vel_x_min'] * t_model_1d
    dfgriddata['pos_y_min'] = dfgriddata['vel_y_min'] * t_model_1d
    dfgriddata['pos_z_min'] = dfgriddata['vel_z_min'] * t_model_1d

    ngridcells = len(dfgriddata['pos_x_min'])
    ncoordgridx = round(ngridcells ** (1. / 3.))
    wid_init = 2 * max(dfgriddata['pos_x_min']) / ncoordgridx
    return dfgriddata, wid_init


def get_cell_midpoints(dfgriddata, wid_init):
    dfgriddata['posx_mid'] = dfgriddata['pos_x_min'] + (0.5 * wid_init)
    dfgriddata['posy_mid'] = dfgriddata['pos_y_min'] + (0.5 * wid_init)
    dfgriddata['posz_mid'] = dfgriddata['pos_z_min'] + (0.5 * wid_init)

    dfgriddata['posx_max'] = dfgriddata['pos_x_min'] + (wid_init)
    dfgriddata['posy_max'] = dfgriddata['pos_y_min'] + (wid_init)
    dfgriddata['posz_max'] = dfgriddata['pos_z_min'] + (wid_init)
    return dfgriddata


def map_1d_to_3d(dfgriddata, vmax, n_3d_gridcells, data_1d, t_model_1d, wid_init):
    modelgridindex = np.zeros(n_3d_gridcells)
    modelgrid_rho_3d = np.zeros(n_3d_gridcells)
    modelgrid_mid_vel = np.zeros(n_3d_gridcells)
    EMPTYGRIDCELL = -99
    vmax_map = 0.25

    for n_3d, row in dfgriddata.iterrows():
        radial_vel_mid = at.vec_len([row['posx_mid']/t_model_1d/CLIGHT,
                                    row['posy_mid']/t_model_1d/CLIGHT, row['posz_mid']/t_model_1d/CLIGHT])

        if n_3d % 1000 == 0:
            print(f'mapping cell {n_3d} of {n_3d_gridcells}')

        if radial_vel_mid < vmax_map:
            for m_1d, radial_vel_1d in enumerate(data_1d['velocity']):
                if radial_vel_1d < radial_vel_mid:

                    modelgridindex[n_3d] = m_1d  # index of outermost 1d cell at 3d cell midpoint
                    modelgrid_rho_3d[n_3d] = data_1d['rho_torus'][m_1d]  # associated rho
                    modelgrid_mid_vel[n_3d] = radial_vel_mid  # associated rho
        else:
            modelgridindex[n_3d] = EMPTYGRIDCELL

    dfgriddata['rho'] = modelgrid_rho_3d
    dfgriddata['vel_mid'] = modelgrid_mid_vel
    # pd.set_option("display.max_rows", None, "display.max_columns", None)
    print(dfgriddata)
    print(sum(modelgrid_rho_3d * (wid_init ** 3))/CLIGHT)

    at.inputmodel.save_modeldata(modelpath='.', dfmodel=dfgriddata, t_model_init_days=t_model_1d/(24*60*60),
                                 dimensions=3, vmax=vmax*CLIGHT)



#!/usr/bin/env python3
"""Artistools.

A collection of plotting, analysis, and file format conversion tools for the ARTIS radiative transfer code.
"""

import sys
from .__main__ import main, addargs
from .commands import commandlist
from .config import config, num_processes, figwidth, enable_diskcache
from .misc import *
from .inputmodel import get_mgi_of_velocity_kms

import artistools.atomic
import artistools.codecomparison
import artistools.deposition
import artistools.estimators
import artistools.inputmodel
import artistools.lightcurve
import artistools.macroatom
import artistools.nltepops
import artistools.nonthermal
import artistools.packets
import artistools.radfield
import artistools.spectra
import artistools.transitions
import artistools.plottools


if sys.version_info < (3,):
    print("Python 2 not supported")



#!/usr/bin/env python3
"""Artistools - light curve functions."""

from .plotlightcurve import main, addargs
from .plotlightcurve import main as plot
from .lightcurve import *
from .viewingangleanalysis import *
